"""
A module which provides utilities for reading, writing, creating and updating
association tables and shift files.

:author: Warren Hack, Nadia Dencheva
:version: '0.1 (2008-01-03)'
"""

from __future__ import absolute_import, division, print_function # confidence high

from . import fileutil as fu
from . import wcsutil
import astropy
from astropy.io import fits
import numpy as N
import os.path, time
from distutils.version import LooseVersion

ASTROPY_VER_GE13 = LooseVersion(astropy.__version__) >= LooseVersion('1.3')

__version__ = '0.2(2015-06-23)'


def readASNTable(fname, output=None, prodonly=False):
    """
    Given a fits filename repesenting an association table reads in the table as a
    dictionary which can be used by pydrizzle and multidrizzle.

    An association table is a FITS binary table with 2 required columns: 'MEMNAME',
    'MEMTYPE'. It checks 'MEMPRSNT' column and removes all files for which its value is 'no'.

    Parameters
    ----------
    fname : str
        name of association table
    output : str
        name of output product - if not specified by the user,
        the first PROD-DTH name is used if present,
        if not, the first PROD-RPT name is used if present,
        if not, the rootname of the input association table is used.
    prodonly : bool
        what files should be considered as input
        if True - select only MEMTYPE=PROD* as input
        if False - select only MEMTYPE=EXP as input

    Returns
    -------
    asndict : dict
        A dictionary-like object with all the association information.

    Examples
    --------
    An association table can be read from a file using the following commands::

    >>> from stsci.tools import asnutil
    >>> asntab = asnutil.readASNTable('j8bt06010_shifts_asn.fits', prodonly=False)  # doctest: +SKIP

    The `asntab` object can now be passed to other code to provide relationships
    between input and output images defined by the association table.

    """

    try:
        f = fits.open(fu.osfn(fname))
    except:
        raise IOError("Can't open file %s\n" % fname)

    colnames = f[1].data.names
    try:
        colunits = f[1].data.units
    except AttributeError: pass

    hdr = f[0].header

    if 'MEMNAME' not in colnames or 'MEMTYPE' not in colnames:
        msg = 'Association table incomplete: required column(s) MEMNAME/MEMTYPE NOT found!'
        raise ValueError(msg)

    d = {}
    for n in colnames:
        d[n]=f[1].data.field(n)
    f.close()

    valid_input = d['MEMPRSNT'].copy()
    memtype = d['MEMTYPE'].copy()
    prod_dth = (memtype.find('PROD-DTH')==0).nonzero()[0]
    prod_rpt = (memtype.find('PROD-RPT')==0).nonzero()[0]
    prod_crj = (memtype.find('PROD-CRJ')==0).nonzero()[0]

    # set output name
    if output is None:
        if prod_dth:
            output = d['MEMNAME'][prod_dth[0]]
        elif prod_rpt:
            output = d['MEMNAME'][prod_rpt[0]]
        elif prod_crj:
            output = d['MEMNAME'][prod_crj[0]]
        else:
            output = fname.split('_')[0]

    if prodonly:
        input = d['MEMTYPE'].find('PROD')==0
        if prod_dth:
            input[prod_dth] = False
    else:
        input = (d['MEMTYPE'].find('EXP')==0)
    valid_input *= input

    for k in d:
        d[k] = d[k][valid_input]

    infiles = list(d['MEMNAME'].lower())
    if not infiles:
        print("No valid input specified")
        return None

    if ('XOFFSET' in colnames and d['XOFFSET'].any()) or ('YOFFSET' in colnames and d['YOFFSET'].any()):
        abshift = True
        dshift = False
        try:
            units=colunits[colnames.index('XOFFSET')]
        except: units='pixels'
        xshifts = list(d['XOFFSET'])
        yshifts = list(d['YOFFSET'])
    elif ('XDELTA' in colnames and d['XDELTA'].any()) or  ('YDELTA' in colnames and d['YDELTA'].any()):
        abshift = False
        dshift = True
        try:
            units=colunits[colnames.index('XDELTA')]
        except: units='pixels'
        xshifts = list(d['XDELTA'])
        yshifts = list(d['YDELTA'])
    else:
        abshift = False
        dshift = False
    members = {}

    if not abshift and not dshift:
        asndict = ASNTable(infiles,output=output)
        asndict.create()
        return asndict
    else:
        try:
            refimage = hdr['refimage']
        except KeyError: refimage = None
        try:
            frame = hdr['shframe']
        except KeyError: frame = 'input'
        if 'ROTATION' in colnames:
            rots = list(d['ROTATION'])
        if 'SCALE' in colnames:
            scales = list(d['SCALE'])

        for r in range(len(infiles)):
            row = r
            xshift = xshifts[r]
            yshift = yshifts[r]
            if rots: rot = rots[r]
            if scales: scale = scales[r]
            members[infiles[r]] = ASNMember(row=row, dshift=dshift, abshift=abshift, rot=rot, xshift=xshift,
                                      yshift=yshift, scale=scale, refimage=refimage, shift_frame=frame,
                                      shift_units=units)


        asndict= ASNTable(infiles, output=output)
        asndict.create()
        asndict['members'].update(members)
        return asndict


class ASNTable(dict):
    """
    A dictionary like object which represents an association table.
    An ASNTable object looks like this::

        {'members':
                {'j8bt06nyq': {'abshift': False,
                           'dshift': True,
                           'refimage': 'j8bt06010_shifts_asn.fits[wcs]',
                           'rot': 0.0,
                           'row': 0,
                           'scale': 1.0,
                           'shift_frame': 'input',
                           'shift_units': 'pixels',
                           'xoff': 0.0,
                           'xshift': 0.0,
                           'yoff': 0.0,
                           'yshift': 0.0},
                'j8bt06nzq': {'abshift': False,
                           'dshift': True,
                           'refimage': 'j8bt06010_shifts_asn.fits[wcs]',
                           'rot': 359.99829,
                           'row': 1,
                           'scale': 1.000165,
                           'shift_frame': 'input',
                           'shift_units': 'pixels',
                           'xoff': 0.0,
                           'xshift': 0.4091132,
                           'yoff': 0.0,
                           'yshift': -0.56702018}},
                'order': ['j8bt06nyq', 'j8bt06nzq'],
                'output': 'j8bt06nyq'}

    Examples
    --------
    Creating an ASNTable object from 3 filenames and a shift file would be done using::

    >>> asnt=ASNTable([fname1,fname2,  fname3], shiftfile='shifts.txt')  # doctest: +SKIP

    The ASNTable object would have the 'members' and 'order'
    in the association table populated based on `infiles` and `shiftfile`.

    This creates a blank association table from the ASNTable object::

    >>> asnt.create()  # doctest: +SKIP

    """
    def __init__(self, inlist=None, output=None, shiftfile=None):
        """
        Parameters
        ----------
        inlist : list
            A list of filenames.
        output :  str
            A user specified output name or 'final'.
        shiftfile : str
            A name of a shift file, if given, the association table will be
            updated with the values in the shift file.

        """

        if output is None:
            if len(inlist) == 1:
                self.output = fu.buildNewRootname(inlist[0])
            else:
                self.output = 'final'
        else:
            self.output = fu.buildNewRootname(output)
            # Ensure that output name does not already contain '_drz'
            _indx = self.output.find('_drz')
            if _indx > 0:
                self.output = self.output[:_indx]

        self.order = []
        if inlist is not None:
            for fn in inlist:
                if fu.findFile(fu.buildRootname(fn)):
                    self.order.append(fu.buildNewRootname(fn))
                else:
                    # This may mean corrupted asn table in which a file is listed as present
                    # when it is missing.
                    raise IOError('File %s not found.\n' %fn)
        dict.__init__(self, output=self.output, order=[], members={})
        if inlist is not None:
            self.input = [fu.buildRootname(f) for f in inlist]
        self.shiftfile = shiftfile

    def create(self, shiftfile=None):
        members = {}
        row = 0
        dshift = False
        abshift = False

        # Parse out shift file, if provided
        if shiftfile is not None:
            sdict = ShiftFile(shiftfile)
        elif self.shiftfile is not None:
            sdict = ShiftFile(self.shiftfile)

            shift_frame = sdict['frame']
            shift_units = sdict['units']
            refimage = sdict['refimage']
            if sdict['form']=='delta':
                dshift = True
            else:
                abshift = True

            for f in self.input:
                xshift = sdict[f][0]
                yshift = sdict[f][1]
                rot = sdict[f][2]
                scale = sdict[f][3]
                #This may not be the right thing to do, may want to keep _flt in rootname
                # to distinguish between _c0h.fits, _c0f.fits and '.c0h'
                fname = fu.buildNewRootname(f)
                members[fname] = ASNMember(row=row, dshift=dshift, abshift=abshift, rot=rot, xshift=xshift,
                                  yshift=yshift, scale=scale, refimage=refimage, shift_frame=shift_frame,
                                  shift_units=shift_units)
                row+=1
        else:
            for f in self.input:
                # also here

                fname = fu.buildNewRootname(f)
                members[fname] = ASNMember(row=row)
                row+=1

        self['members'].update(members)
        self['order']=self.order


    def update(self, members=None, shiftfile=None, replace=False):
        __help_update="""
        Update an existing association table.

        Parameters
        ----------
        members : dict
            A dictionary representing asndict['members'].
        shiftfile : str
            The name of a shift file
            If given, shiftfile will replace shifts in an asndict.
        replace : bool False(default)
            A flag which indicates whether the 'members' item
            of an association table should be updated or replaced.
            default: False
            If True, it's up to the user to replace also asndict['order']
        """
        if members and isinstance(members, dict):
            if not replace:
                self['members'].update(members=members)
            else:
                self['members'] = members
        elif shiftfile:
            members = {}
            abshift = False
            dshift = False
            row = 0
            sdict = ShiftFile(shiftfile)
            shift_frame = sdict['frame']
            shift_units = sdict['units']
            refimage = sdict['refimage']
            if sdict['form']=='delta':
                dshift = True
            else:
                abshift = True

            for f in self.order:
                fullname = fu.buildRootname(f)
                xshift = sdict[fullname][0]
                yshift = sdict[fullname][1]
                rot = sdict[fullname][2]
                scale = sdict[fullname][3]
                members[f] = ASNMember(row=row, dshift=dshift, abshift=abshift, rot=rot, xshift=xshift,
                                  yshift=yshift, scale=scale, refimage=refimage, shift_frame=shift_frame,
                                  shift_units=shift_units)
                row+=1
            self['members'].update(members)
        else:
            #print __help_update
            pass

    def write(self, output=None):
        """
        Write association table to a file.

        """
        if not output:
            outfile = self['output']+'_asn.fits'
            output = self['output']
        else:
            outfile = output

        # Delete the file if it exists.
        if os.path.exists(outfile):
            warningmsg =  "\n#########################################\n"
            warningmsg += "#                                       #\n"
            warningmsg += "# WARNING:                              #\n"
            warningmsg += "#  The existing association table,      #\n"
            warningmsg += "           " + str(outfile) + '\n'
            warningmsg += "#  is being replaced.                   #\n"
            warningmsg += "#                                       #\n"
            warningmsg += "#########################################\n\n"
        fasn = fits.HDUList()

        # Compute maximum length of MEMNAME for table column definition
        _maxlen = 0
        for _fname in self['order']:
            if len(_fname) > _maxlen: _maxlen = len(_fname)
        # Enforce a mimimum size of 24
        if _maxlen < 24: _maxlen = 24
        namelen_str = str(_maxlen+2)+'A'
        self.buildPrimary(fasn, output=output)

        mname = self['order'][:]
        mname.append(output)
        mtype = ['EXP-DTH' for l in self['order']]
        mtype.append('PROD-DTH')
        mprsn = [True for l in self['order']]
        mprsn.append(False)
        xoff = [self['members'][l]['xoff'] for l in self['order']]
        xoff.append(0.0)
        yoff = [self['members'][l]['yoff'] for l in self['order']]
        yoff.append(0.0)
        xsh = [self['members'][l]['xshift'] for l in self['order']]
        xsh.append(0.0)
        ysh = [self['members'][l]['yshift'] for l in self['order']]
        ysh.append(0.0)
        rot = [self['members'][l]['rot'] for l in self['order']]
        rot.append(0.0)
        scl = [self['members'][l]['scale'] for l in self['order']]
        scl.append(1.0)

        memname = fits.Column(name='MEMNAME',format=namelen_str,array=N.char.array(mname))
        memtype = fits.Column(name='MEMTYPE',format='14A',array=N.char.array(mtype))
        memprsn = fits.Column(name='MEMPRSNT', format='L', array=N.array(mprsn).astype(N.uint8))
        xoffset = fits.Column(name='XOFFSET', format='E', array=N.array(xoff))
        yoffset = fits.Column(name='YOFFSET', format='E', array=N.array(yoff))
        xdelta = fits.Column(name='XDELTA', format='E', array=N.array(xsh))
        ydelta = fits.Column(name='YDELTA', format='E', array=N.array(ysh))
        rotation = fits.Column(name='ROTATION', format='E', array=N.array(rot))
        scale = fits.Column(name='SCALE', format='E', array=N.array(scl))
        cols = fits.ColDefs([memname,memtype,memprsn,xoffset,yoffset,xdelta,ydelta,rotation,scale])
        hdu = fits.BinTableHDU.from_columns(cols)
        fasn.append(hdu)
        if ASTROPY_VER_GE13:
            fasn.writeto(outfile, overwrite=True)
        else:
            fasn.writeto(outfile, clobber=True)
        fasn.close()
        mem0 = self['order'][0]
        refimg = self['members'][mem0]['refimage']
        if refimg is not None:
            whdu = wcsutil.WCSObject(refimg)
            whdu.createReferenceWCS(outfile,overwrite=False)
            ftab = fits.open(outfile)
            ftab['primary'].header['refimage'] = outfile+"[wcs]"
            ftab.close()
        del whdu



    def buildPrimary(self, fasn, output=None):
        _prihdr = fits.Header([fits.Card('SIMPLE', True, 'Fits standard'),
                    fits.Card('BITPIX  ',                    16 ,' Bits per pixel'),
                    fits.Card('NAXIS   ',                     0 ,' Number of axes'),
                    fits.Card('ORIGIN  ',  'NOAO-IRAF FITS Image Kernel July 1999' ,'FITS file originator'),
                    fits.Card('IRAF-TLM',  '18:26:13 (27/03/2000)' ,' Time of last modification'),
                    fits.Card('EXTEND  ', True ,' File may contain standard extensions'),
                    fits.Card('NEXTEND ',                     1 ,' Number of standard extensions'),
                    fits.Card('DATE    ',  '2001-02-14T20:07:57',' date this file was written (yyyy-mm-dd)'),
                    fits.Card('FILENAME',  'hr_box_asn.fits'            ,' name of file'),
                    fits.Card('FILETYPE',  'ASN_TABLE'          ,' type of data found in data file'),
                    fits.Card('TELESCOP',  'HST'                ,' telescope used to acquire data'),
                    fits.Card('INSTRUME',  'ACS   '             ,' identifier for instrument used to acquire data'),
                    fits.Card('EQUINOX ',                2000.0 ,' equinox of celestial coord. system'),
                    fits.Card('ROOTNAME',  'hr_box  '              ,' rootname of the observation set'),
                    fits.Card('PRIMESI ',  'ACS   '             ,' instrument designated as prime'),
                    fits.Card('TARGNAME',  'SIM-DITHER'                     ,'proposer\'s target name'),
                    fits.Card('RA_TARG ',                    0. ,' right ascension of the target (deg) (J2000)'),
                    fits.Card('DEC_TARG',                    0. ,' declination of the target (deg) (J2000)'),
                    fits.Card('DETECTOR',  'HRC     '           ,' detector in use: WFC, HRC, or SBC'),
                    fits.Card('ASN_ID  ',  'hr_box  '           ,' unique identifier assigned to association'),
                    fits.Card('ASN_TAB ',  'hr_box_asn.fits'         ,' name of the association table')])

        # Format time values for keywords IRAF-TLM, and DATE
        _ltime = time.localtime(time.time())
        tlm_str = time.strftime('%H:%M:%S (%d/%m/%Y)',_ltime)
        date_str = time.strftime('%Y-%m-%dT%H:%M:%S',_ltime)
        origin_str = 'FITS Version '+ astropy.__version__
        # Build PRIMARY HDU
        _hdu = fits.PrimaryHDU(header=_prihdr)
        fasn.append(_hdu)

        newhdr = fasn['PRIMARY'].header
        mem0name = self['order'][0]
        refimg = self['members'][mem0name]['refimage']
        shframe = self['members'][mem0name]['shift_frame']
        fullname = fu.buildRootname(mem0name,ext=['_flt.fits', '_c0h.fits', '_c0f.fits'])
        try:
            # Open img1 to obtain keyword values for updating template
            fimg1 = fits.open(fullname)
        except:
            print('File %s does not exist' % fullname)


        kws = ['INSTRUME', 'PRIMESI', 'TARGNAME', 'DETECTOR', 'RA_TARG', 'DEC_TARG']
        mem0hdr = fimg1['PRIMARY'].header
        default = 'UNKNOWN'
        for kw in kws:
            try:
                newhdr[kw] = mem0hdr[kw]
            except:
                newhdr[kw] = default
        fimg1.close()

        if not output:
            output = self['output']

        outfilename = fu.buildNewRootname(output, extn='_asn.fits')
        newhdr['IRAF-TLM']=tlm_str
        newhdr['DATE'] = date_str
        newhdr['ORIGIN'] = origin_str
        newhdr['ROOTNAME'] = output

        newhdr['FILENAME'] = outfilename
        newhdr['ASN_ID'] = output
        newhdr['ASN_TAB'] = outfilename
        newhdr['SHFRAME'] = (shframe, "Frame which shifts are measured")
        newhdr['REFIMAGE'] = (refimg, "Image shifts were measured from")



class ASNMember(dict):
    """
    A dictionary like object representing a member of an association table. It looks like this::

        'j8bt06nzq': {'abshift': False,
                  'dshift': True,
                  'refimage': 'j8bt06010_shifts_asn.fits[wcs]',
                  'rot': 359.99829,
                  'row': 1,
                  'scale': 1.000165,
                  'shift_frame': 'input',
                  'shift_units': 'pixels',
                  'xoff': 0.0,
                  'xshift': 0.4091132,
                  'yoff': 0.0,
                  'yshift': -0.56702018}

    If `abshift` is True, shifts, rotation and scale refer to absolute shifts.
    If `dshift`  is True, they are delta shifts.

    """

    def __init__(self, xoff=0.0, yoff=0.0, rot=0.0, xshift=0.0,
                 yshift=0.0, scale=1.0, dshift=False, abshift=False, refimage="", shift_frame="",
                 shift_units='pixels', row=0):

        dict.__init__(self, xoff=xoff, yoff=yoff, xshift=xshift, yshift=yshift, rot=rot, scale=scale,
                      dshift=dshift, abshift=abshift, refimage=refimage, shift_frame=shift_frame,
                      shift_units=shift_units, row=row)

class ShiftFile(dict):
    """
    A shift file has the following format (name, Xsh, Ysh, Rot, Scale)::

        # frame: output
        # refimage: tweak_wcs.fits[wcs]
        # form: delta
        # units: pixels
        j8bt06nyq_flt.fits    0.0  0.0    0.0    1.0
        j8bt06nzq_flt.fits    0.4091132  -0.5670202    359.9983    1.000165

    This object creates a `dict` like object representing a shift file used by Pydrizzle and Mirashift.
    """

    def __init__(self,filename="", form='delta', frame=None, units='pixels',
                 order=None, refimage=None, **kw):
        """
        :Purpose: Create a dict like ShiftFile object from a shift file on disk or from
                  variables in memory. If a file name is provided all other parameters are ignored.

        Examples
        ---------
        These examples demonstrate a couple of the most common usages.

        Read a shift file on disk using::

        >>> sdict = ShiftFile('shifts.txt')  # doctest: +SKIP

        Pass values for the fields of the shift file and a dictionary with all files::

        >>> d={'j8bt06nyq_flt.fits': [0.0, 0.0, 0.0, 1.0],
        ...    'j8bt06nzq_flt.fits': [0.4091132, -0.5670202, 359.9983, 1.000165]}

        >>> sdict = ShiftFile(
        ...     form='absolute', frame='output', units='pixels',
        ...     order=['j8bt06nyq_flt.fits', 'j8bt06nzq_flt.fits'],
        ...     refimage='tweak_wcs.fits[wcs]', **d)  # doctest: +SKIP

        The return value can then be used to provide the shift information to code in memory.

        Parameters
        ----------
        filename : str
            Name of shift file on disk, see above the expected format
        form : str
            Form of shifts (absolute|delta)
        frame : str
            Frame in which the shifts should be applied (input|output)
        units : str
            Units in which the shifts are measured.
        order : list
            Keeps track of the order of the files.
        refimage : str
                    name of reference image
         **d :  dict
                    keys: file names
                    values: a list:  [Xsh, Ysh, Rot, Scale]
                    The keys must match the files in the order parameter.

        Raises
        ------
        ValueError
            If reference file can't be found

        """
        ## History: This is refactored code which was initially in fileutil.py and
        ## pydrizzle: buildasn.py and updateasn.py

        dict.__init__(self, form=form, frame=frame, units=units,order=order, refimage=refimage)

        if filename == "":
            self.update(kw)
        else:
            self.readShiftFile(filename)

        if not self.verifyShiftFile():
            msg = "\nReference image not found.\n "
            msg += "The keyword in the shift file has changed from 'reference' to 'refimage'.\n"
            msg += "Make sure this keyword is specified as 'refimage' in %s." %filename

            raise ValueError(msg)

    def readShiftFile(self, filename):
        """
        Reads a shift file from disk and populates a dictionary.
        """
        order = []
        fshift = open(filename,'r')
        flines = fshift.readlines()
        fshift.close()

        common = [f.strip('#').strip() for f in flines if f.startswith('#')]
        c=[line.split(': ') for line in common]

        # Remove any line comments in the shift file - lines starting with '#'
        # but not part of the common block.
        for l in c:
            if l[0] not in ['frame', 'refimage', 'form', 'units']:
                c.remove(l)

        for line in c: line[1]=line[1].strip()
        self.update(c)

        files = [f.strip().split(' ',1) for f in flines if not (f.startswith('#') or f.strip() == '')]
        for f in files:
            order.append(f[0])

        self['order'] = order

        for f in files:
            # Check to see if filename provided is a full filename that corresponds
            # to a file on the path.  If not, try to convert given rootname into
            # a valid filename based on available files.  This may or may not
            # define the correct filename, which is why it prints out what it is
            # doing, so that the user can verify and edit the shiftfile if needed.
            #NOTE:
            # Supporting the specification of only rootnames in the shiftfile with this
            # filename expansion is NOT to be documented, but provided solely as
            # an undocumented, dangerous and not fully supported helper function for
            # some backwards compatibility.
            if not os.path.exists(f[0]):
                f[0] = fu.buildRootname(f[0])
                print('Defining filename in shiftfile as: ', f[0])

            f[1] = f[1].split()
            try:
                f[1] = [float(s) for s in f[1]]
            except:
                msg = 'Cannot read in ', s, ' from shiftfile ', filename, ' as a float number'
                raise ValueError(msg)
            msg = "At least 2 and at most 4 shift values should be provided in a shiftfile"
            if len(f[1]) < 2:
                raise ValueError(msg)
            elif len(f[1]) == 3:
                f[1].append(1.0)
            elif len(f[1]) == 2:
                f[1].extend([0.0, 1.0])
            elif len(f[1]) > 4:
                raise ValueError(msg)

        fdict = dict(files)
        self.update(fdict)


    def verifyShiftFile(self):
        """
        Verifies that reference file exists.
        """
        if self['refimage'] and fu.findFile(self['refimage']):
            return True
        else: return False

    def writeShiftFile(self, filename="shifts.txt"):
        """
        Writes a shift file object to a file on disk using the convention for shift file format.
        """
        lines = ['# frame: ', self['frame'], '\n',
                 '# refimage: ', self['refimage'], '\n',
                 '# form: ', self['form'], '\n',
                 '# units: ', self['units'], '\n']

        for o in self['order']:
            ss = " "
            for shift in self[o]:
                ss += str(shift) + " "
            line = str(o) + ss + "\n"
            lines.append(line)

        fshifts= open(filename, 'w')
        fshifts.writelines(lines)
        fshifts.close()



from __future__ import absolute_import

import os
import tempfile

import pytest
from astropy.io.fits.tests import FitsTestCase

from .. import fileutil as F


class TestIsFits(FitsTestCase):
    def setup(self):
        self.data_dir = os.path.join(os.path.dirname(__file__), 'data')
        self.temp_dir = tempfile.mkdtemp(prefix='isfits-test-')

    def test_isFits(self):
        assert F.isFits(self.data('cdva2.fits')) == (True, 'simple')
        assert F.isFits(self.data('o4sp040b0_raw.fits')) == (True, 'mef')
        assert F.isFits(self.data('waivered.fits')) == (True, 'waiver')

        # isFits only verifies files with valid FITS extensions (.fits,...)
        # Does not matter if file does not exist.
        assert F.isFits(self.data('simple')) == (False, None)

        # But if it has FITS extension, should raise error if nonexistent.
        with pytest.raises(IOError):
            F.isFits('isfits/no_such_file.fits')



""" Read-Only tkinter Text Widget.  This is a variation of the tkinter Text
widget in that the text itself is not editable (it is read-only), but it allows
selection for cut/paste to other apps.  Cut-paste may currently only work
under X11. (9/2015 enabled under OSX by adding 'c' to ALLOWED_SYMS)

A vastly simpler way of doing this is to use a tkinter.Text widget and set
it to DISABLED, but then you cannot select text.
$Id$
"""
from __future__ import division  # confidence high

# System level modules
import sys
PY3K = sys.version_info[0] > 2

if PY3K:
    import tkinter as TKNTR
else:
    import Tkinter as TKNTR

ALLOWED_SYMS = ('c', 'Up', 'Down', 'Left', 'Right', 'Home', 'End', 'Prior',
                'Next', 'Shift_L', 'Shift_R')


class ROText(TKNTR.Text):

    def __init__(self, master, **kw):
        """  defer most of __init__ to the base class """
        self._fbto = None
        if 'focusBackTo' in kw:
            self._fbto = kw['focusBackTo']
            del kw['focusBackTo']
        TKNTR.Text.__init__(self, master, **kw)
        # override some bindings to return a "break" string
        self.bind("<Key>", self.ignoreMostKeys)
        self.bind("<Button-2>", lambda e: "break")
        self.bind("<Button-3>", lambda e: "break")
        if self._fbto:
            self.bind("<Leave>", self.mouseLeft)
        self.config(insertwidth=0)

    # disallow common insert calls, but allow a backdoor when needed
    def insert(self, index, text, *tags, **kw):
        if 'force' in kw and kw['force']:
            TKNTR.Text.insert(self, index, text, *tags)

    # disallow common delete calls, but allow a backdoor when needed
    def delete(self, start, end=None, force=False):
        if force:
            TKNTR.Text.delete(self, start, end)

    # a way to disable text manip
    def ignoreMostKeys(self, event):
        if event.keysym not in ALLOWED_SYMS:
            return "break"  # have to return this string to stop the event
        # To get copy/paste working on OSX we allow 'c' so that
        # they can type 'Command-c', but don't let a regular 'c' through.
        if event.keysym in ('c', 'C'):
            if (sys.platform == 'darwin' and hasattr(event, 'state') and
                    event.state != 0):
                pass  # allow this through, it is Command-c
            else:
                return "break"

    def mouseLeft(self, event):
        if self._fbto:
            self._fbto.focus_set()
        return "break"  # have to return this string to stop the event



# $Id$

"""
The stpyfits module is an extension to the `astropy.io.fits` module which offers
additional features specific to STScI.  These features include the handling
of Constant Data Value Arrays.

"""
from __future__ import division

import functools
import sys
import numpy as np

import astropy
from astropy.io import fits
# A few imports for backward compatibility; in the earlier stpyfits these were
# overridden, but with fits's new extension system it's not necessary
from astropy.io.fits.util import _is_int
from astropy.utils import lazyproperty
from distutils.version import LooseVersion

PY3K = sys.version_info[0] > 2
ASTROPY_VER_GE20 = LooseVersion(astropy.__version__) >= LooseVersion('2.0')

STPYFITS_ENABLED = False  # Not threadsafe TODO: (should it be?)


# Register the extension classes; simply importing stpyfits does not
# automatically enable it.  Instead, it can be enabled/disabled using these
# functions.
def enable_stpyfits():
    global STPYFITS_ENABLED
    if not STPYFITS_ENABLED:
        fits.register_hdu(ConstantValuePrimaryHDU)
        fits.register_hdu(ConstantValueImageHDU)
        STPYFITS_ENABLED = True


def disable_stpyfits():
    global STPYFITS_ENABLED
    if STPYFITS_ENABLED:
        fits.unregister_hdu(ConstantValuePrimaryHDU)
        fits.unregister_hdu(ConstantValueImageHDU)
        STPYFITS_ENABLED = False


def with_stpyfits(func):
    @functools.wraps(func)
    def wrapped_with_stpyfits(*args, **kwargs):
        global STPYFITS_ENABLED
        was_enabled = STPYFITS_ENABLED
        enable_stpyfits()
        try:
            # BUG: Forcefully disable lazy loading.
            # Lazy loading breaks ability to initialize ConstantValueHDUs
            # TODO: Investigate the cause upstream (astropy.io.fits)
            if 'write' not in func.__name__:
                kwargs['lazy_load_hdus'] = False
            retval = func(*args, **kwargs)
        finally:
            # Only disable stpyfits if it wasn't already enabled
            if not was_enabled:
                disable_stpyfits()
        return retval
    return wrapped_with_stpyfits


class _ConstantValueImageBaseHDU(fits.hdu.image._ImageBaseHDU):
    """
    A class that extends the `astropy.io.fits.hdu.base._BaseHDU` class to extend its
    behavior to implement STScI specific extensions to `astropy.io.fits`.

    The `astropy.io.fits.hdu.base._BaseHDU class` is:
    """

    __doc__ += fits.hdu.image._ImageBaseHDU.__doc__

    def __init__(self, data=None, header=None, do_not_scale_image_data=False,
                 uint=False, **kwargs):
        if header and 'PIXVALUE' in header and header['NAXIS'] == 0:
            header = header.copy()
            # Add NAXISn keywords for each NPIXn keyword in the header and
            # remove the NPIXn keywords
            naxis = 0
            for card in reversed(header['NPIX*'].cards):
                try:
                    idx = int(card.keyword[len('NPIX'):])
                except ValueError:
                    continue
                hdrlen = len(header)
                header.set('NAXIS' + str(idx), card.value,
                           card.comment, after='NAXIS')
                del header[card.keyword]
                if len(header) < hdrlen:
                    # A blank card was used when updating the header; add the
                    # blank back in.
                    # TODO: Fix header.set so that it has an option not to
                    # use a blank card--this is a detail that we really
                    # shouldn't have to worry about otherwise
                    header.append()

                # Presumably the NPIX keywords are in order of their axis, but
                # just in case somehow they're not...
                naxis = max(naxis, idx)

            # Update the NAXIS keyword with the correct number of axes
            header['NAXIS'] = naxis
        elif header and 'PIXVALUE' in header:
            pixval = header['PIXVALUE']
            if header['BITPIX'] > 0:
                if PY3K:
                    pixval = int(pixval)
                else:
                    pixval = long(pixval)
            arrayval = self._check_constant_value_data(data)
            if arrayval is not None:
                header = header.copy()
                # Update the PIXVALUE keyword if necessary
                if arrayval != pixval:
                    header['PIXVALUE'] = arrayval
            else:
                header = header.copy()
                # There is a PIXVALUE keyword but NAXIS is not 0 and the data
                # does not match the PIXVALUE.
                # Must remove the PIXVALUE and NPIXn keywords so we recognize
                # that there is non-constant data in the file.
                del header['PIXVALUE']
                for card in header['NPIX*'].cards:
                    try:
                        idx = int(card.keyword[len('NPIX'):])
                    except ValueError:
                        continue
                    del header[card.keyword]

        # Make sure to pass any arguments other than data and header as
        # keyword arguments, because PrimaryHDU and ImageHDU have stupidly
        # different signatures for __init__
        super(_ConstantValueImageBaseHDU, self).__init__(
            data, header, do_not_scale_image_data=do_not_scale_image_data,
            uint=uint)

    @property
    def size(self):
        """
        The HDU's size should always come up as zero so long as there's no
        actual data in it other than the constant value array.
        """

        if 'PIXVALUE' in self._header:
            return 0
        else:
            return super(_ConstantValueImageBaseHDU, self).size

    @lazyproperty
    def data(self):
        if ('PIXVALUE' in self._header and 'NPIX1' not in self._header and
                self._header['NAXIS'] > 0):
            bitpix = self._header['BITPIX']
            dims = self.shape

            # Special case where the pixvalue can be present but all the NPIXn
            # keywords are zero.
            if sum(dims) == 0:
                return None

            code = BITPIX2DTYPE[bitpix]
            pixval = self._header['PIXVALUE']
            if code in ['uint8', 'int16', 'int32', 'int64']:
                if PY3K:
                    pixval = int(pixval)
                else:
                    pixval = long(pixval)

            raw_data = np.zeros(shape=dims, dtype=code) + pixval

            if raw_data.dtype.str[0] != '>':
                raw_data = raw_data.byteswap(True)

            raw_data.dtype = raw_data.dtype.newbyteorder('>')

            if self._bzero != 0 or self._bscale != 1:
                if bitpix > 16:  # scale integers to Float64
                    data = np.array(raw_data, dtype=np.float64)
                elif bitpix > 0:  # scale integers to Float32
                    data = np.array(raw_data, dtype=np.float32)
                else:  # floating point cases
                    data = raw_data

                if self._bscale != 1:
                    np.multiply(data, self._bscale, data)
                if self._bzero != 0:
                    data += self._bzero

                # delete the keywords BSCALE and BZERO after scaling
                del self._header['BSCALE']
                del self._header['BZERO']
                self._header['BITPIX'] = DTYPE2BITPIX[data.dtype.name]
            else:
                data = raw_data
            return data
        else:
            return super(_ConstantValueImageBaseHDU, self).data

    @data.setter
    def data(self, data):
        self.__dict__['data'] = data
        self._modified = True
        if self.data is not None and not isinstance(data, np.ndarray):
            # Try to coerce the data into a numpy array--this will work, on
            # some level, for most objects
            try:
                data = np.array(data)
            except:
                raise TypeError('data object %r could not be coerced into an '
                                'ndarray' % data)

        if isinstance(data, np.ndarray):
            self._bitpix = DTYPE2BITPIX[data.dtype.name]
            self._axes = list(data.shape)
            self._axes.reverse()
        elif self.data is None:
            self._axes = []
        else:
            raise ValueError('not a valid data array')

        self.update_header()

    @classmethod
    def match_header(cls, header):
        """A constant value HDU will only be recognized as such if the header
        contains a valid PIXVALUE and NAXIS == 0.
        """

        pixvalue = header.get('PIXVALUE')
        naxis = header.get('NAXIS', 0)

        return (super(_ConstantValueImageBaseHDU, cls).match_header(header) and
                (isinstance(pixvalue, float) or _is_int(pixvalue)) and
                naxis == 0)

    def update_header(self):
        if (not self._modified and not self._header._modified and
                (self._has_data and self.shape == self.data.shape)):
            # Not likely that anything needs updating
            return

        super(_ConstantValueImageBaseHDU, self).update_header()

        if 'PIXVALUE' in self._header and self._header['NAXIS'] > 0:
            # This is a Constant Value Data Array.  Verify that the data
            # actually matches the PIXVALUE
            pixval = self._header['PIXVALUE']
            if self._header['BITPIX'] > 0:
                if PY3K:
                    pixval = int(pixval)
                else:
                    pixval = long(pixval)

            if self.data is None or self.data.nbytes == 0:
                # Empty data array; just keep the existing PIXVALUE
                arrayval = self._header['PIXVALUE']
            else:
                arrayval = self._check_constant_value_data(self.data)
            if arrayval is not None:
                if arrayval != pixval:
                    self._header['PIXVALUE'] = arrayval

                naxis = self._header['NAXIS']
                self._header['NAXIS'] = 0
                for idx in range(naxis, 0, -1):
                    axisval = self._header['NAXIS%d' % idx]
                    self._header.set('NPIX%d' % idx, axisval,
                                     'length of constant array axis %d' % idx,
                                     after='PIXVALUE')
                    del self._header['NAXIS%d' % idx]
            else:
                # No longer a constant value array; remove any remaining
                # NPIX or PIXVALUE keywords
                try:
                    del self._header['PIXVALUE']
                except KeyError:
                    pass

                try:
                    del self._header['NPIX*']
                except KeyError:
                    pass

    def _summary(self):
        summ = super(_ConstantValueImageBaseHDU, self)._summary()
        if ASTROPY_VER_GE20:
            outsumm = ((summ[0], summ[1],
                        summ[2].replace('ConstantValue', '')) + summ[3:])
        else:
            outsumm = ((summ[0],
                        summ[1].replace('ConstantValue', '')) + summ[2:])
        return outsumm

    def _writedata_internal(self, fileobj):
        if 'PIXVALUE' in self._header:
            # This is a Constant Value Data Array, so no data is written
            return 0
        else:
            return super(_ConstantValueImageBaseHDU, self)._writedata_internal(fileobj)

    def _check_constant_value_data(self, data):
        """Verify that the HDU's data is a constant value array."""

        arrayval = data.flat[0]
        if np.all(data == arrayval):
            return arrayval
        return None


class ConstantValuePrimaryHDU(_ConstantValueImageBaseHDU,
                              fits.hdu.PrimaryHDU):
    """Primary HDUs with constant value arrays."""


class ConstantValueImageHDU(_ConstantValueImageBaseHDU, fits.hdu.ImageHDU):
    """Image extension HDUs with constant value arrays."""


# Import the rest of the astropy.io.fits module
from astropy.io.fits import *  # noqa

# For backward-compatibility with older code that thinks PrimaryHDU and
# ImageHDU should support the ConstantValue features
PrimaryHDU = ConstantValuePrimaryHDU
ImageHDU = ConstantValueImageHDU


# Override the other "convenience" functions to use stpyfits
open = fitsopen = with_stpyfits(fits.open)
info = with_stpyfits(fits.info)
append = with_stpyfits(fits.append)
writeto = with_stpyfits(fits.writeto)
update = with_stpyfits(fits.update)
getheader = with_stpyfits(fits.getheader)
getdata = with_stpyfits(fits.getdata)
getval = with_stpyfits(fits.getval)
setval = with_stpyfits(fits.setval)
delval = with_stpyfits(fits.delval)


__all__ = fits.__all__ + ['enable_stpyfits', 'disable_stpyfits',
                          'with_stpyfits', 'ConstantValuePrimaryHDU',
                          'ConstantValueImageHDU']



"""fitsdiff is now a part of Astropy.

Now this module just provides a wrapper around astropy.io.fits.diff for backwards
compatibility with the old interface in case anyone uses it.
"""
import os
import sys

from astropy.io.fits.diff import FITSDiff
from astropy.io.fits.scripts.fitsdiff import log, main

PY3K = sys.version_info[0] > 2
if PY3K:
    string_types = str
else:
    string_types = basestring


def fitsdiff(input1, input2, comment_excl_list='', value_excl_list='',
             field_excl_list='', maxdiff=10, delta=0.0, neglect_blanks=True,
             output=None):

    if isinstance(comment_excl_list, string_types):
        comment_excl_list = list_parse(comment_excl_list)

    if isinstance(value_excl_list, string_types):
        value_excl_list = list_parse(value_excl_list)

    if isinstance(field_excl_list, string_types):
        field_excl_list = list_parse(field_excl_list)

    diff = FITSDiff(input1, input2, ignore_keywords=value_excl_list,
                    ignore_comments=comment_excl_list,
                    ignore_fields=field_excl_list, numdiffs=maxdiff,
                    tolerance=delta, ignore_blanks=neglect_blanks)

    if output is None:
        output = sys.stdout

    diff.report(output)

    return diff.identical


def list_parse(name_list):
    """Parse a comma-separated list of values, or a filename (starting with @)
    containing a list value on each line.
    """

    if name_list and name_list[0] == '@':
        value = name_list[1:]
        if not os.path.exists(value):
            log.warning('The file %s does not exist' % value)
            return
        try:
            return [v.strip() for v in open(value, 'r').readlines()]
        except IOError as e:
            log.warning('reading %s failed: %s; ignoring this file' %
                        (value, e))
    else:
        return [v.strip() for v in name_list.split(',')]


if __name__ == "__main__":
    sys.exit(main())



from __future__ import absolute_import

import sys
import time

import pytest

from ..tkrotext import ROText

PY3K = sys.version_info[0] > 2

if PY3K:
    import tkinter as TKNTR
else:
    import Tkinter as TKNTR


def test_rotest():
    rot = None

    def quit():
        with pytest.raises(SystemExit):
            sys.exit()

    def clicked():
        rot.insert(TKNTR.END, "\nClicked at " + time.asctime(), force=True)
        rot.see(TKNTR.END)

    # make our test window
    try:
        top = TKNTR.Tk()
    except Exception as e:
        pytest.xfail(str(e))  # Travis does not have interactive session
    f = TKNTR.Frame(top)

    sc = TKNTR.Scrollbar(f)
    sc.pack(side=TKNTR.RIGHT, fill=TKNTR.Y)
    rot = ROText(f, wrap=TKNTR.WORD, height=10, yscrollcommand=sc.set,
                 focusBackTo=top)
    rot.pack(side=TKNTR.TOP, fill=TKNTR.X, expand=True)
    sc.config(command=rot.yview)
    f.pack(side=TKNTR.TOP, fill=TKNTR.X)

    b = TKNTR.Button(top, text='Click Me', command=clicked)
    b.pack(side=TKNTR.TOP, fill=TKNTR.X, expand=1)

    q = TKNTR.Button(top, text='Quit', command=quit)
    q.pack(side=TKNTR.TOP)

    # start
    top.mainloop()



"""
:Module: xyinterp.py

Interpolates y based on the given xval.

`x` and `y` are a pair of independent/dependent variable arrays that must
be the same length. The x array must also be sorted.
`xval` is a user-specified value. This routine looks
up `xval` in the x array and uses that information to properly interpolate
the value in the y array. 


:author: Vicki Laidler

:version: '0.1 (2006-07-06)'  


"""
from __future__ import division # confidence high
import numpy as N 

#This section for standalone imports only-------------------------------------
__version__ = '0.1'          #Release version number only
__vdate__ = '2006-07-06'     #Date of this version, in this (FITS-style) format
#-----------------------------------------------------------------------------


def xyinterp(x,y,xval):
    """ 
    
    :Purpose: Interpolates y based on the given xval.

    x and y are a pair of independent/dependent variable arrays that must
    be the same length. The x array must also be sorted.
    xval is a user-specified value. This routine looks
    up xval in the x array and uses that information to properly interpolate
    the value in the y array.  

    Notes
    =====
    Use the searchsorted method on the X array to determine the bin in
    which xval falls; then use that information to compute the corresponding
    y value.
    

    See Also 
    ========
    numpy

    Parameters
    ==========

    x: 1D numpy array  
        independent variable array: MUST BE SORTED

    y: 1D numpy array
        dependent variable array

    xval: float 
        the x value at which you want to know the value of y

    Returns
    =======
    y: float 
        the value of y corresponding to xval

    Raises
    ======
    ValueError: 
        If arrays are unequal length; or x array is unsorted;
        or if xval falls outside the bounds of x (extrapolation is unsupported

    :version: 0.1 last modified 2006-07-06

"""

    #Enforce conditions on x, y, and xval:
    #x and y must correspond
    if len(x) != len(y):
        raise ValueError("Input arrays must be equal lengths")

    #Extrapolation not supported
    if xval < x[0]:
        raise ValueError("Value %f < min(x) %f: Extrapolation unsupported"%(xval,x[0]))
    if xval > x[-1]:
        raise ValueError("Value > max(x): Extrapolation unsupported")

    #This algorithm only works on sorted data
    if x.argsort().all() != N.arange(len(x)).all():
        raise ValueError("Input array x must be sorted")
    
    # Now do the real work.
    hi = x.searchsorted(xval)
    lo = hi - 1
    
    try:
        seg = (float(xval)-x[lo]) / (x[hi] - x[lo])
    except ZeroDivisionError:
        seg = 0.0

    yval = y[lo] + seg*(y[hi] - y[lo])
    return yval



"""
License: http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE
"""
from __future__ import absolute_import, division # confidence high
import glob
try:
    from .fileutil import osfn # added to provide interpretation of environment variables
except:
    osfn = None
__author__ = 'Paul Barrett'
__version__ = '1.1'

def irafglob(inlist, atfile=None):
    """ Returns a list of filenames based on the type of IRAF input.

    Handles lists, wild-card characters, and at-files.  For special
    at-files, use the atfile keyword to process them.

    This function is recursive, so IRAF lists can also contain at-files
    and wild-card characters, e.g. `a.fits`, `@file.lst`, `*flt.fits`.
    """

    # Sanity check
    if inlist is None or len(inlist) == 0:
        return []

    # Determine which form of input was provided:
    if isinstance(inlist, list):
        #  python list
        flist = []
        for f in inlist:
            flist += irafglob(f)
    elif ',' in inlist:
        #  comma-separated string list
        flist = []
        for f in inlist.split(','):
            f = f.strip()
            flist += irafglob(f)
    elif inlist[0] == '@':
        #  file list
        flist = []
        for f in open(inlist[1:], 'r').readlines():
            f = f.rstrip()
            # hook for application specific atfiles.
            if atfile:
                f = atfile(f)
            flist += irafglob(f)
    else:
        #  shell globbing
        if osfn:
            inlist = osfn(inlist)
        flist = glob.glob(inlist)

    return flist



#!/usr/bin/env python
#
"""
   This module is from Lennart Regebro's ComparableMixin class, available at:

       http://regebro.wordpress.com/2010/12/13/
              python-implementing-rich-comparison-the-correct-way/

   The idea is to prevent you from having to define lt,le,eq,ne,etc...
   This may no longer be necessary after the functools total_ordering
   decorator (Python v2.7) is available on all Python versions
   supported by our software.

   For simple comparisons, all that is necessary is to derive your class
   from ComparableMixin and override the _cmpkey() method.

   For more complex comparisons (where type-checking needs to occur and
   comparisons to other types are allowed), simply override _compare() instead
   of _cmpkey().

   BEWARE that comparing different types has different results in Python 2.x
   versus Python 3.x:

        Python 2.7
        >>> 'a' < 2  # doctest: +SKIP
        False

        Python 3.2.1
        >>> 'a' < 2   # doctest: +SKIP
        Traceback (most recent call last):
          File "<stdin>", line 1, in <module>
        TypeError: unorderable types: str() < int()
"""
from __future__ import print_function


class ComparableMixin(object):
    def _compare(self, other, method):
        try:
            return method(self._cmpkey(), other._cmpkey())
        except (AttributeError, TypeError):
            # _cmpkey not implemented, or return different type,
            # so I can't compare with "other".
            return NotImplemented

    def __lt__(self, other):
        return self._compare(other, lambda s,o: s < o)

    def __le__(self, other):
        return self._compare(other, lambda s,o: s <= o)

    def __eq__(self, other):
        return self._compare(other, lambda s,o: s == o)

    def __ge__(self, other):
        return self._compare(other, lambda s,o: s >= o)

    def __gt__(self, other):
        return self._compare(other, lambda s,o: s > o)

    def __ne__(self, other):
        return self._compare(other, lambda s,o: s != o)


class ComparableIntBaseMixin(ComparableMixin):
    """ For those classes which, at heart, are comparable to integers. """
    def _compare(self, other, method):
        if isinstance(other, self.__class__): # two objects of same class
            return method(self._cmpkey(), other._cmpkey())
        else:
            return method(int(self._cmpkey()), int(other))


class ComparableFloatBaseMixin(ComparableMixin):
    """ For those classes which, at heart, are comparable to floats. """
    def _compare(self, other, method):
        if isinstance(other, self.__class__): # two objects of same class
            return method(self._cmpkey(), other._cmpkey())
        else:
            return method(float(self._cmpkey()), float(other))



""" Contains the ConfigObjPars class and any related functionality.

$Id$
"""
from __future__ import absolute_import, division, print_function # confidence high

import copy, glob, os, stat, sys

if sys.version_info[0] > 2:
    string_types = str
else:
    string_types = basestring

# ConfigObj modules
from . import configobj, validate

# Local modules
from . import basicpar, eparoption, irafutils, taskpars, vtor_checks

# Globals and useful functions

APP_NAME = 'TEAL'
TASK_NAME_KEY = '_task_name_'

class DuplicateKeyError(Exception):
    pass

class NoCfgFileError(Exception):
    pass


def getAppDir():
    """ Return our application dir.  Create it if it doesn't exist. """
    # Be sure the resource dir exists
    theDir = os.path.expanduser('~/.')+APP_NAME.lower()
    if not os.path.exists(theDir):
        try:
            os.mkdir(theDir)
        except OSError:
            print('Could not create "'+theDir+'" to save GUI settings.')
            theDir = "./"+APP_NAME.lower()
    return theDir


def getObjectFromTaskArg(theTask, strict, setAllToDefaults):
    """ Take the arg (usually called theTask), which can be either a subclass
    of ConfigObjPars, or a string package name, or a .cfg filename - no matter
    what it is - take it and return a ConfigObjPars object.
    strict - bool - warning severity, passed to the ConfigObjPars() ctor
    setAllToDefaults - bool - if theTask is a pkg name, force all to defaults
    """

    # Already in the form we need (instance of us or of subclass)
    if isinstance(theTask, ConfigObjPars):
        if setAllToDefaults:
            raise RuntimeError('Called getObjectFromTaskArg with existing'+\
                  ' object AND setAllToDefaults - is unexpected use case.')
        # If it is an existing object, make sure it's internal param list is
        # up to date with it's ConfigObj dict, since the user may have manually
        # edited the dict before calling us.
        theTask.syncParamList(False) # use strict somehow?
        # Note - some validation is done here in IrafPar creation, but it is
        # not the same validation done by the ConfigObj s/w (no check funcs).
        # Do we want to do that too here?
        return theTask

    # For example, a .cfg file
    if os.path.isfile(str(theTask)):
        try:
            return ConfigObjPars(theTask, strict=strict,
                                 setAllToDefaults=setAllToDefaults)
        except KeyError:
            # this might just be caused by a file sitting in the local cwd with
            # the same exact name as the package we want to import, let's see
            if theTask.find('.') > 0: # it has an extension, like '.cfg'
                raise # this really was an error
            # else we drop down to the next step - try it as a pkg name

    # Else it must be a Python package name to load
    if isinstance(theTask, str) and setAllToDefaults:
        # NOTE how we pass the task name string in setAllToDefaults
        return ConfigObjPars('', setAllToDefaults=theTask, strict=strict)
    else:
        return getParsObjForPyPkg(theTask, strict)


def getEmbeddedKeyVal(cfgFileName, kwdName, dflt=None):
    """ Read a config file and pull out the value of a given keyword. """
    # Assume this is a ConfigObj file.  Use that s/w to quickly read it and
    # put it in dict format.  Assume kwd is at top level (not in a section).
    # The input may also be a .cfgspc file.
    #
    # Only use ConfigObj here as a tool to generate a dict from a file - do
    # not use the returned object as a ConfigObj per se.  As such, we can call
    # with "simple" format, ie. no cfgspc, no val'n, and "list_values"=False.
    try:
        junkObj = configobj.ConfigObj(cfgFileName, list_values=False)
    except:
        if kwdName == TASK_NAME_KEY:
            raise KeyError('Can not parse as a parameter config file: '+ \
                           '\n\t'+os.path.realpath(cfgFileName))
        else:
            raise KeyError('Unfound key "'+kwdName+'" while parsing: '+ \
                           '\n\t'+os.path.realpath(cfgFileName))

    if kwdName in junkObj:
        retval = junkObj[kwdName]
        del junkObj
        return retval
    # Not found
    if dflt is not None:
        del junkObj
        return dflt
    else:
        if kwdName == TASK_NAME_KEY:
            raise KeyError('Can not parse as a parameter config file: '+ \
                           '\n\t'+os.path.realpath(cfgFileName))
        else:
            raise KeyError('Unfound key "'+kwdName+'" while parsing: '+ \
                           '\n\t'+os.path.realpath(cfgFileName))


def findCfgFileForPkg(pkgName, theExt, pkgObj=None, taskName=None):
    """ Locate the configuration files for/from/within a given python package.
    pkgName is a string python package name.  This is used unless pkgObj
    is given, in which case pkgName is taken from pkgObj.__name__.
    theExt is either '.cfg' or '.cfgspc'. If the task name is known, it is
    given as taskName, otherwise one is determined using the pkgName.
    Returns a tuple of (package-object, cfg-file-name). """
    # arg check
    ext = theExt
    if ext[0] != '.': ext = '.'+theExt

    # Do the import, if needed
    pkgsToTry = {}
    if pkgObj:
        pkgsToTry[pkgObj.__name__] = pkgObj
    else:
        # First try something simple like a regular or dotted import
        try:
            fl = []
            if pkgName.find('.') > 0:
                fl = [ pkgName[:pkgName.rfind('.')], ]
            pkgsToTry[str(pkgName)] = __import__(str(pkgName), fromlist=fl)
        except:
            throwIt = True
            # One last case to try is something like "csc_kill" from
            # "acstools.csc_kill", but this convenience capability will only be
            # allowed if the parent pkg (acstools) has already been imported.
            if isinstance(pkgName, string_types) and pkgName.find('.') < 0:
                matches = [x for x in sys.modules.keys() \
                           if x.endswith("."+pkgName)]
                if len(matches)>0:
                    throwIt = False
                    for mmm in matches:
                        pkgsToTry[mmm] = sys.modules[mmm]
            if throwIt:
                raise NoCfgFileError("Unfound package or "+ext+" file via: "+\
                                     "import "+str(pkgName))

    # Now that we have the package object (or a few of them to try), for each
    # one find the .cfg or .cfgspc file, and return
    # Return as soon as ANY match is found.
    for aPkgName in pkgsToTry:
        aPkg = pkgsToTry[aPkgName]
        path = os.path.dirname(aPkg.__file__)
        if len(path) < 1: path = '.'
        flist = irafutils.rglob(path, "*"+ext)
        if len(flist) < 1:
            continue

        # Go through these and find the first one for the assumed or given task
        # name.  The task name for 'BigBlackBox.drizzle' would be 'drizzle'.
        if taskName is None:
            taskName = aPkgName.split(".")[-1]
        flist.sort()
        for f in flist:
            # A .cfg file gets checked for _task_name_=val, but a .cfgspc file
            # will have a string check function signature as the val.
            if ext == '.cfg':
                itsTask = getEmbeddedKeyVal(f, TASK_NAME_KEY, '')
            else: # .cfgspc
                sigStr  = getEmbeddedKeyVal(f, TASK_NAME_KEY, '')
                # .cfgspc file MUST have an entry for TASK_NAME_KEY w/ a default
                itsTask = vtor_checks.sigStrToKwArgsDict(sigStr)['default']
            if itsTask == taskName:
                # We've found the correct file in an installation area.  Return
                # the package object and the found file.
                return aPkg, f

    # What, are you still here?
    raise NoCfgFileError('No valid '+ext+' files found in package: "'+ \
                         str(pkgName)+'" for task: "'+str(taskName)+'"')


def findAllCfgTasksUnderDir(aDir):
    """ Finds all installed tasks by examining any .cfg files found on disk
        at and under the given directory, as an installation might be.
        This returns a dict of { file name : task name }
    """
    retval = {}
    for f in irafutils.rglob(aDir, '*.cfg'):
        retval[f] = getEmbeddedKeyVal(f, TASK_NAME_KEY, '')
    return retval


def getCfgFilesInDirForTask(aDir, aTask, recurse=False):
    """ This is a specialized function which is meant only to keep the
        same code from needlessly being much repeated throughout this
        application.  This must be kept as fast and as light as possible.
        This checks a given directory for .cfg files matching a given
        task.  If recurse is True, it will check subdirectories.
        If aTask is None, it returns all files and ignores aTask.
    """
    if recurse:
        flist = irafutils.rglob(aDir, '*.cfg')
    else:
        flist = glob.glob(aDir+os.sep+'*.cfg')
    if aTask:
        retval = []
        for f in flist:
            try:
                if aTask == getEmbeddedKeyVal(f, TASK_NAME_KEY, ''):
                    retval.append(f)
            except Exception as e:
                print('Warning: '+str(e))
        return retval
    else:
        return flist


def getParsObjForPyPkg(pkgName, strict):
    """ Locate the appropriate ConfigObjPars (or subclass) within the given
        package. NOTE this begins the same way as getUsrCfgFilesForPyPkg().
        Look for .cfg file matches in these places, in this order:
          1 - any named .cfg file in current directory matching given task
          2 - if there exists a ~/.teal/<taskname>.cfg file
          3 - any named .cfg file in SOME*ENV*VAR directory matching given task
          4 - the installed default .cfg file (with the given package)
    """
    # Get the python package and it's .cfg file - need this no matter what
    installedPkg, installedFile = findCfgFileForPkg(pkgName, '.cfg')
    theFile = None
    tname = getEmbeddedKeyVal(installedFile, TASK_NAME_KEY)

    # See if the user has any of their own .cfg files in the cwd for this task
    if theFile is None:
        flist = getCfgFilesInDirForTask(os.getcwd(), tname)
        if len(flist) > 0:
            if len(flist) == 1: # can skip file times sort
                theFile = flist[0]
            else:
                # There are a few different choices.  In the absence of
                # requirements to the contrary, just take the latest.  Set up a
                # list of tuples of (mtime, fname) so we can sort by mtime.
                ftups = [ (os.stat(f)[stat.ST_MTIME], f) for f in flist]
                ftups.sort()
                theFile = ftups[-1][1]

    # See if the user has any of their own app-dir .cfg files for this task
    if theFile is None:
        flist = getCfgFilesInDirForTask(getAppDir(), tname) # verifies tname
        flist = [f for f in flist if os.path.basename(f) == tname+'.cfg']
        if len(flist) > 0:
            theFile = flist[0]
            assert len(flist) == 1, str(flist) # should never happen

    # Add code to check an env. var defined area?  (speak to users first)

    # Did we find one yet?  If not, use the installed version
    useInstVer = False
    if theFile is None:
        theFile = installedFile
        useInstVer = True

    # Create a stand-in instance from this file.  Force a read-only situation
    # if we are dealing with the installed, (expected to be) unwritable file.
    return ConfigObjPars(theFile, associatedPkg=installedPkg,
                         forceReadOnly=useInstVer, strict=strict)


def getUsrCfgFilesForPyPkg(pkgName):
    """ See if the user has one of their own local .cfg files for this task,
        such as might be created automatically during the save of a read-only
        package, and return their names. """
    # Get the python package and it's .cfg file
    thePkg, theFile = findCfgFileForPkg(pkgName, '.cfg')
    # See if the user has any of their own local .cfg files for this task
    tname = getEmbeddedKeyVal(theFile, TASK_NAME_KEY)
    flist = getCfgFilesInDirForTask(getAppDir(), tname)
    return flist


def checkSetReadOnly(fname, raiseOnErr = False):
    """ See if we have write-privileges to this file.  If we do, and we
    are not supposed to, then fix that case. """
    if os.access(fname, os.W_OK):
        # We can write to this but it is supposed to be read-only. Fix it.
        # Take away usr-write, leave group and other alone, though it
        # may be simpler to just force/set it to: r--r--r-- or r--------
        irafutils.setWritePrivs(fname, False, ignoreErrors= not raiseOnErr)


def flattenDictTree(aDict):
    """ Takes a dict of vals and dicts (so, a tree) as input, and returns
    a flat dict (only one level) as output.  All key-vals are moved to
    the top level.  Sub-section dict names (keys) are ignored/dropped.
    If there are name collisions, an error is raised. """
    retval = {}
    for k in aDict:
        val = aDict[k]
        if isinstance(val, dict):
            # This val is a dict, get its data (recursively) into a flat dict
            subDict = flattenDictTree(val)
            # Merge its dict of data into ours, watching for NO collisions
            rvKeySet  = set(retval.keys())
            sdKeySet = set(subDict.keys())
            intr = rvKeySet.intersection(sdKeySet)
            if len(intr) > 0:
                raise DuplicateKeyError("Flattened dict already has "+ \
                    "key(s): "+str(list(intr))+" - cannot flatten this.")

            else:
                retval.update(subDict)
        else:
            if k in retval:
                raise DuplicateKeyError("Flattened dict already has key: "+\
                                        k+" - cannot flatten this.")
            else:
                retval[k] = val
    return retval


def countKey(theDict, name):
    """ Return the number of times the given par exists in this dict-tree,
    since the same key name may be used in different sections/sub-sections. """

    retval = 0
    for key in theDict:
        val = theDict[key]
        if isinstance(val, dict):
            retval += countKey(val, name) # recurse
        else:
            if key == name:
                retval += 1
                # can't break, even tho we found a hit, other items on
                # this level will not be named "name", but child dicts
                # may have further counts
    return retval


def findFirstPar(theDict, name, _depth=0):
    """ Find the given par.  Return tuple: (its own (sub-)dict, its value).
    Returns the first match found, without checking whether the given key name
    is unique or whether it is used in multiple sections. """

    for key in theDict:
        val = theDict[key]
#       print _depth*'   ', key, str(val)[:40]
        if isinstance(val, dict):
            retval = findFirstPar(val, name, _depth=_depth+1) # recurse
            if retval is not None:
                return retval
            # else keep looking
        else:
            if key == name:
                return theDict, theDict[name]
            # else keep looking
    # if we get here then we have searched this whole (sub)-section and its
    # descendants, and found no matches.  only raise if we are at the top.
    if _depth == 0:
        raise KeyError(name)
    else:
        return None


def findScopedPar(theDict, scope, name):
    """ Find the given par.  Return tuple: (its own (sub-)dict, its value). """
    # Do not search (like findFirstPar), but go right to the correct
    # sub-section, and pick it up.  Assume it is there as stated.
    if len(scope):
        theDict = theDict[scope] # ! only goes one level deep - enhance !
    return theDict, theDict[name] # KeyError if unfound


def setPar(theDict, name, value):
    """ Sets a par's value without having to give its scope/section. """
    section, previousVal = findFirstPar(theDict, name)
    # "section" is the actual object, not a copy
    section[name] = value


def mergeConfigObj(configObj, inputDict):
    """ Merge the inputDict values into an existing given configObj instance.
    The inputDict is a "flat" dict - it has no sections/sub-sections.  The
    configObj may have sub-sections nested to any depth.  This will raise a
    DuplicateKeyError if one of the inputDict keys is used more than once in
    configObj (e.g. within two different sub-sections). """
    # Expanded upon Warren's version in astrodrizzle

    # Verify that all inputDict keys in configObj are unique within configObj
    for key in inputDict:
        if countKey(configObj, key) > 1:
            raise DuplicateKeyError(key)
    # Now update configObj with each inputDict item
    for key in inputDict:
        setPar(configObj, key, inputDict[key])


def integrityTestAllPkgCfgFiles(pkgObj, output=True):
    """ Given a package OBJECT, inspect it and find all installed .cfg file-
    using tasks under it.  Then them one at a time via
    integrityTestTaskCfgFile, and report any/all errors. """
    assert type(pkgObj) == type(os), \
           "Expected module arg, got: "+str(type(pkgObj))
    taskDict = findAllCfgTasksUnderDir(os.path.dirname(pkgObj.__file__))
    # taskDict is { cfgFileName : taskName }
    errors = []
    for fname in taskDict:
        taskName = taskDict[fname]
        try:
            if taskName:
                if output:
                    print('In '+pkgObj.__name__+', checking task: '+
                           taskName+', file: '+fname)
                integrityTestTaskCfgFile(taskName, fname)
        except Exception as e:
            errors.append(str(e))

    assert len(errors) == 0, 'Errors found while integrity testing .cfg '+ \
              'file(s) found under "'+pkgObj.__name__+'":\n'+ \
              ('\n'.join(errors))


def integrityTestTaskCfgFile(taskName, cfgFileName=None):
    """ For a given task, inspect the given .cfg file (or simply find/use its
    installed .cfg file), and check those values against the defaults
    found in the installed .cfgspc file.  They should be the same.
    If the file name is not given, the installed one is found and used. """

    from . import teal # don't import above, to avoid circular import (may need to mv)
    if not cfgFileName:
        ignored, cfgFileName = findCfgFileForPkg(taskName, '.cfg')
    diffDict = teal.diffFromDefaults(cfgFileName, report=False)
    if len(diffDict) < 1:
        return # no error
    msg = 'The following par:value pairs from "'+cfgFileName+ \
          '" are not the correct defaults: '+str(diffDict)
    raise RuntimeError(msg)


class ConfigObjPars(taskpars.TaskPars, configobj.ConfigObj):
    """ This represents a task's dict of ConfigObj parameters. """

    def __init__(self, cfgFileName, forUseWithEpar=True,
                 setAllToDefaults=False, strict=True,
                 associatedPkg=None, forceReadOnly=False):
        """
        cfgFileName - string path/name of .cfg file
        forUseWithEpar - bool - will this be used in EPAR?
        setAllToDefaults - <True, False, or string> string is pkg name to import
        strict - bool - level of error/warning severity
        associatedPkg - loaded package object
        forceReadOnly - bool - make the .cfg file read-only
        """

        self._forUseWithEpar = forUseWithEpar
        self._rcDir = getAppDir()
        self._allTriggers = None # all known triggers in this object
        self._allDepdcs = None   # all known dependencies in this object
        self._allExecutes = None # all known codes-to-execute in this object
        self._neverWrite = []    # all keys which are NOT written out to .cfg
        self._debugLogger = None
        self._debugYetToPost = []
        self.__assocPkg = associatedPkg

        # The __paramList pointer remains the same for the life of this object
        self.__paramList = []

        # Set up ConfigObj stuff
        assert setAllToDefaults or os.path.isfile(cfgFileName), \
               "Config file not found: "+cfgFileName
        self.__taskName = ''
        if setAllToDefaults:
            # they may not have given us a real file name here since they
            # just want defaults (in .cfgspc) so don't be too picky about
            # finding and reading the file.
            if isinstance(setAllToDefaults, str):
                # here they have very clearly said to load only the defaults
                # using the given name as the package name - below we will
                # have it imported in _findAssociatedConfigSpecFile()
                self.__taskName = setAllToDefaults
                setAllToDefaults = True
                cfgFileName = '' # ignore any given .cfg file, don't need one
            else:
                possible = os.path.splitext(os.path.basename(cfgFileName))[0]
                if os.path.isfile(cfgFileName):
                    self.__taskName = getEmbeddedKeyVal(cfgFileName,
                                      TASK_NAME_KEY, possible)
                else:
                    self.__taskName = possible
        else:
            # this is the real deal, expect a real file name
            self.__taskName = getEmbeddedKeyVal(cfgFileName, TASK_NAME_KEY)
            if forceReadOnly:
                checkSetReadOnly(cfgFileName)

        # Find the associated .cfgspc file (first make sure we weren't
        # given one by mistake)
        if not cfgFileName.endswith('.cfg') and \
           self.__taskName.find('(default=') >= 0:
            # Handle case where they gave us a .cfgspc by mistake (no .cfg)
            # (basically reset a few things)
            cfgSpecPath = os.path.realpath(cfgFileName)
            setAllToDefaults = True
            cfgFileName = ''
            sigStr  = getEmbeddedKeyVal(cfgSpecPath, TASK_NAME_KEY, '')
            self.__taskName = vtor_checks.sigStrToKwArgsDict(sigStr)['default']
        else:
            cfgSpecPath = self._findAssociatedConfigSpecFile(cfgFileName)
        assert os.path.exists(cfgSpecPath), \
               "Matching configspec not found!  Expected: "+cfgSpecPath

        self.debug('ConfigObjPars: .cfg='+str(cfgFileName)+ \
                   ', .cfgspc='+str(cfgSpecPath)+ \
                   ', defaults='+str(setAllToDefaults)+', strict='+str(strict))

        # Run the ConfigObj ctor.  The result of this (if !setAllToDefaults)
        # is the exact copy of the input file as a dict (ConfigObj).  If the
        # infile had extra pars or missing pars, they are still that way here.
        if setAllToDefaults:
            configobj.ConfigObj.__init__(self, configspec=cfgSpecPath)
        else:
            configobj.ConfigObj.__init__(self, os.path.abspath(cfgFileName),
                                         configspec=cfgSpecPath)

        # Before we validate (and fill in missing pars), find any lost pars
        # via this (somewhat kludgy) method suggested by ConfigObj folks.
        missing = '' # assume no .cfg file
        if strict and (not setAllToDefaults):
            # don't even populate this if not strict
            missing = findTheLost(os.path.abspath(cfgFileName), cfgSpecPath)

        # Validate it here.  We can't skip this step even if we are just
        # setting all to defaults, since this sets the values.
        # NOTE - this fills in values for any missing pars !  AND, if our
        # .cfgspc sets defaults vals, then missing pars are not an error...
        self._vtor = validate.Validator(vtor_checks.FUNC_DICT)
        # 'ans' will be True, False, or a dict (anything but True is bad)
        ans = self.validate(self._vtor, preserve_errors=True,
                            copy=setAllToDefaults)
        # Note: before the call to validate(), the list returned from
        # self.keys() is in the order found in self.filename.  If that file
        # was missing items that are in the .cfgspc, they will now show up
        # in self.keys(), but not necessarily in the same order as the .cfgspc
        hasTypeErr = ans != True
        extra = self.listTheExtras(True)

        # DEAL WITH ERRORS (in this way)
        #
        # wrong par type:
        #     strict -> severe error*
        #     not -> severe error
        # extra par(s) found:
        #     strict -> severe error
        #     not -> warn*
        # missing par(s):
        #     strict -> warn
        #     not - be silent
        #
        # *severe - if in GUI, pop up error & stop (e.g. file load), else raise
        # *warn - if in GUI, pop up warning, else print it to screen

        if extra or missing or hasTypeErr:
            flatStr = ''
            if ans == False:
                flatStr = "All values are invalid!"
            if ans != True and ans != False:
                flatStr = flattened2str(configobj.flatten_errors(self, ans))
            if missing:
                flatStr += "\n\n"+missing
            if extra:
                flatStr += "\n\n"+extra
            msg = "Validation warnings for: "
            if hasTypeErr or (strict and extra):
                msg = "Validation errors for: "
            msg = msg+os.path.realpath(cfgFileName)+\
                  "\n\n"+flatStr.strip('\n')
            if hasTypeErr or (strict and extra):
                raise RuntimeError(msg)
            else:
                # just inform them, but don't throw anything
                print(msg.replace('\n\n','\n'))

        # get the initial param list out of the ConfigObj dict
        self.syncParamList(True)

        # take note of all trigger logic
        self.debug(self.triggerLogicToStr())

        # see if we are using a package with it's own run() function
        self._runFunc = None
        self._helpFunc = None
        if self.__assocPkg is not None:
            if hasattr(self.__assocPkg, 'run'):
                self._runFunc = self.__assocPkg.run
            if hasattr(self.__assocPkg, 'getHelpAsString'):
                self._helpFunc = self.__assocPkg.getHelpAsString


    def setDebugLogger(self, obj):
        # set the object we can use to post debugging info
        self._debugLogger = obj
        # now that we have one, post anything we have saved up (and clear list)
        if obj and len(self._debugYetToPost) > 0:
            for msg in self._debugYetToPost:
                self._debugLogger.debug(msg)
        self._debugYetToPost = []

    def debug(self, msg):
        if self._debugLogger:
            self._debugLogger.debug(msg)
        else:
            # else just hold onto it until we do have a logger -during the
            # init phase we may not yet have a logger, yet have stuff to log
            self._debugYetToPost.append(msg) # add to our little cache

    def getDefaultSaveFilename(self, stub=False):
        """ Return name of file where we are expected to be saved if no files
        for this task have ever been saved, and the user wishes to save.  If
        stub is True, the result will be <dir>/<taskname>_stub.cfg instead of
        <dir>/<taskname>.cfg. """
        if stub:
            return self._rcDir+os.sep+self.__taskName+'_stub.cfg'
        else:
            return self._rcDir+os.sep+self.__taskName+'.cfg'

    def syncParamList(self, firstTime, preserve_order=True):
        """ Set or reset the internal param list from the dict's contents. """
        # See the note in setParam about this design.

        # Get latest par values from dict.  Make sure we do not
        # change the id of the __paramList pointer here.
        new_list = self._getParamsFromConfigDict(self, initialPass=firstTime)
                                               # dumpCfgspcTo=sys.stdout)
        # Have to add this odd last one for the sake of the GUI (still?)
        if self._forUseWithEpar:
            new_list.append(basicpar.IrafParS(['$nargs','s','h','N']))

        if len(self.__paramList) > 0 and preserve_order:
            # Here we have the most up-to-date data from the actual data
            # model, the ConfigObj dict, and we need to use it to fill in
            # our param list.  BUT, we need to preserve the order our list
            # has had up until now (by unique parameter name).
            namesInOrder = [p.fullName() for p in self.__paramList]
            assert len(namesInOrder) == len(new_list), \
                   'Mismatch in num pars, had: '+str(len(namesInOrder))+ \
                   ', now we have: '+str(len(new_list))+', '+ \
                   str([p.fullName() for p in new_list])
            self.__paramList[:] = [] # clear list, keep same pointer
            # create a flat dict view of new_list, for ease of use in next step
            new_list_dict = {} # can do in one step in v2.7
            for par in new_list: new_list_dict[par.fullName()] = par
            # populate
            for fn in namesInOrder:
                self.__paramList.append(new_list_dict[fn])
        else:
            # Here we just take the data in whatever order it came.
            self.__paramList[:] = new_list # keep same list pointer

    def getName(self): return self.__taskName

    def getPkgname(self):  return '' # subclasses override w/ a sensible value

    def getParList(self, docopy=False):
        """ Return a list of parameter objects.  docopy is ignored as the
        returned value is not a copy. """
        return self.__paramList

    def getDefaultParList(self):
        """ Return a par list just like ours, but with all default values. """
        # The code below (create a new set-to-dflts obj) is correct, but it
        # adds a tenth of a second to startup.  Clicking "Defaults" in the
        # GUI does not call this.  But this can be used to set the order seen.

        # But first check for rare case of no cfg file name
        if self.filename is None:
            # this is a .cfgspc-only kind of object so far
            self.filename = self.getDefaultSaveFilename(stub=True)
            return copy.deepcopy(self.__paramList)

        tmpObj = ConfigObjPars(self.filename, associatedPkg=self.__assocPkg,
                               setAllToDefaults=True, strict=False)
        return tmpObj.getParList()

    def getFilename(self):
        if self.filename in (None, ''):
            return self.getDefaultSaveFilename()
        else:
            return self.filename

    def getAssocPkg(self):
        return self.__assocPkg

    def canExecute(self):
        return self._runFunc is not None

    def isSameTaskAs(self, aCfgObjPrs):
        """ Return True if the passed in object is for the same task as
        we are. """
        return aCfgObjPrs.getName() == self.getName()

#   def strictUpdate(self, aDict):
#       """ Override the current values with those in the given dict.  This
#           is like dict's update, except it doesn't allow new keys and it
#           verifies the values (it does?!) """
#       if aDict is None:
#           return
#       for k in aDict:
#           v = aDict[k]
#           print("Skipping ovverride key = "+k+", val = "+str(v))

    def setParam(self, name, val, scope='', check=1, idxHint=None):
        """ Find the ConfigObj entry.  Update the __paramList. """
        theDict, oldVal = findScopedPar(self, scope, name)

        # Set the value, even if invalid.  It needs to be set before
        # the validation step (next).
        theDict[name] = val

        # If need be, check the proposed value.  Ideally, we'd like to
        # (somehow elegantly) only check this one item. For now, the best
        # shortcut is to only validate this section.
        if check:
            ans=self.validate(self._vtor, preserve_errors=True, section=theDict)
            if ans != True:
                flatStr = "All values are invalid!"
                if ans != False:
                    flatStr = flattened2str(configobj.flatten_errors(self, ans))
                raise RuntimeError("Validation error: "+flatStr)

        # Note - this design needs work.  Right now there are two copies
        # of the data:  the ConfigObj dict, and the __paramList ...
        # We rely on the idxHint arg so we don't have to search the __paramList
        # every time this is called, which could really slows things down.
        assert idxHint is not None, "ConfigObjPars relies on a valid idxHint"
        assert name == self.__paramList[idxHint].name, \
               'Error in setParam, name: "'+name+'" != name at idxHint: "'+\
               self.__paramList[idxHint].name+'", idxHint: '+str(idxHint)
        self.__paramList[idxHint].set(val)

    def saveParList(self, *args, **kw):
        """Write parameter data to filename (string or filehandle)"""
        if 'filename' in kw:
            filename = kw['filename']
        if not filename:
            filename = self.getFilename()
        if not filename:
            raise ValueError("No filename specified to save parameters")

        if hasattr(filename,'write'):
            fh = filename
            absFileName = os.path.abspath(fh.name)
        else:
            absFileName = os.path.expanduser(filename)
            absDir = os.path.dirname(absFileName)
            if len(absDir) and not os.path.isdir(absDir): os.makedirs(absDir)
            fh = open(absFileName,'w')
        numpars = len(self.__paramList)
        if self._forUseWithEpar: numpars -= 1
        if not self.final_comment: self.final_comment = [''] # force \n at EOF
        # Empty the ConfigObj version of section.defaults since that is based
        # on an assumption incorrect for us, and override with our own list.
        # THIS IS A BIT OF MONKEY-PATCHING!  WATCH FUTURE VERSION CHANGES!
        # See Trac ticket #762.
        while len(self.defaults):
            self.defaults.pop(-1) # empty it, keeping ref
        for key in self._neverWrite:
            self.defaults.append(key)
        # Note also that we are only overwriting the top/main section's
        # "defaults" list, but EVERY [sub-]section has such an attribute...

        # Now write to file, delegating work to ConfigObj (note that ConfigObj
        # write() skips any items listed by name in the self.defaults list)
        self.write(fh)
        fh.close()
        retval = str(numpars) + " parameters written to " + absFileName
        self.filename = absFileName # reset our own ConfigObj filename attr
        self.debug('Keys not written: '+str(self.defaults))
        return retval

    def run(self, *args, **kw):
        """ This may be overridden by a subclass. """
        if self._runFunc is not None:
            # remove the two args sent by EditParDialog which we do not use
            if 'mode' in kw: kw.pop('mode')
            if '_save' in kw: kw.pop('_save')
            return self._runFunc(self, *args, **kw)
        else:
            raise taskpars.NoExecError('No way to run task "'+self.__taskName+\
                '". You must either override the "run" method in your '+ \
                'ConfigObjPars subclass, or you must supply a "run" '+ \
                'function in your package.')

    def triggerLogicToStr(self):
        """ Print all the trigger logic to a string and return it. """
        try:
            import json
        except ImportError:
            return "Cannot dump triggers/dependencies/executes (need json)"
        retval = "TRIGGERS:\n"+json.dumps(self._allTriggers, indent=3)
        retval += "\nDEPENDENCIES:\n"+json.dumps(self._allDepdcs, indent=3)
        retval += "\nTO EXECUTE:\n"+json.dumps(self._allExecutes, indent=3)
        retval += "\n"
        return retval


    def getHelpAsString(self):
        """ This may be overridden by a subclass. """
        if self._helpFunc is not None:
            return self._helpFunc()
        else:
            return 'No help string found for task "'+self.__taskName+ \
            '".  \n\nThe developer must either override the '+\
            'getHelpAsString() method in their ConfigObjPars \n'+ \
            'subclass, or they must supply such a function in their package.'

    def _findAssociatedConfigSpecFile(self, cfgFileName):
        """ Given a config file, find its associated config-spec file, and
        return the full pathname of the file. """

        # Handle simplest 2 cases first: co-located or local .cfgspc file
        retval = "."+os.sep+self.__taskName+".cfgspc"
        if os.path.isfile(retval): return retval

        retval = os.path.dirname(cfgFileName)+os.sep+self.__taskName+".cfgspc"
        if os.path.isfile(retval): return retval

        # Also try the resource dir
        retval = self.getDefaultSaveFilename()+'spc' # .cfgspc
        if os.path.isfile(retval): return retval

        # Now try and see if there is a matching .cfgspc file in/under an
        # associated package, if one is defined.
        if self.__assocPkg is not None:
            x, theFile = findCfgFileForPkg(None, '.cfgspc',
                                           pkgObj = self.__assocPkg,
                                           taskName = self.__taskName)
            return theFile

        # Finally try to import the task name and see if there is a .cfgspc
        # file in that directory
        x, theFile = findCfgFileForPkg(self.__taskName, '.cfgspc',
                                       taskName = self.__taskName)
        if os.path.exists(theFile):
            return theFile

        # unfound
        raise NoCfgFileError('Unfound config-spec file for task: "'+ \
                             self.__taskName+'"')


    def _getParamsFromConfigDict(self, cfgObj, scopePrefix='',
                                 initialPass=False, dumpCfgspcTo=None):
        """ Walk the given ConfigObj dict pulling out IRAF-like parameters into
        a list. Since this operates on a dict this can be called recursively.
        This is also our chance to find and pull out triggers and such
        dependencies. """
        # init
        retval = []
        if initialPass and len(scopePrefix) < 1:
            self._posArgs = [] # positional args [2-tuples]: (index,scopedName)
            # FOR SECURITY: the following 3 chunks of data,
            #     _allTriggers, _allDepdcs, _allExecutes,
            # are collected ONLY from the .cfgspc file
            self._allTriggers = {}
            self._allDepdcs = {}
            self._allExecutes = {}

        # start walking ("tell yer story walkin, buddy")
        # NOTE: this relies on the "in" operator returning keys in the
        # order that they exist in the dict (which depends on ConfigObj keeping
        # the order they were found in the original file)
        for key in cfgObj:
            val = cfgObj[key]

            # Do we need to skip this - if not a par, like a rule or something
            toBeHidden = isHiddenName(key)
            if toBeHidden:
                if key not in self._neverWrite and key != TASK_NAME_KEY:
                    self._neverWrite.append(key)
                    # yes TASK_NAME_KEY is hidden, but it IS output to the .cfg

            # a section
            if isinstance(val, dict):
                if not toBeHidden:
                    if len(list(val.keys()))>0 and len(retval)>0:
                        # Here is where we sneak in the section comment
                        # This is so incredibly kludgy (as the code was), it
                        # MUST be revamped eventually! This is for the epar GUI.
                        prevPar = retval[-1]
                        # Use the key (or its comment?) as the section header
                        prevPar.set(prevPar.get('p_prompt')+'\n\n'+key,
                                    field='p_prompt', check=0)
                    if dumpCfgspcTo:
                        dumpCfgspcTo.write('\n['+key+']\n')
                    # a logical grouping (append its params)
                    pfx = scopePrefix+'.'+key
                    pfx = pfx.strip('.')
                    retval = retval + self._getParamsFromConfigDict(val, pfx,
                                      initialPass, dumpCfgspcTo) # recurse
            else:
                # a param
                fields = []
                choicesOrMin = None
                fields.append(key) # name
                dtype = 's'
                cspc = None
                if cfgObj.configspec:
                    cspc = cfgObj.configspec.get(key) # None if not found
                chk_func_name = ''
                chk_args_dict = {}
                if cspc:
                    chk_func_name = cspc[:cspc.find('(')]
                    chk_args_dict = vtor_checks.sigStrToKwArgsDict(cspc)
                if chk_func_name.find('option') >= 0:
                    dtype = 's'
                    # convert the choices string to a list (to weed out kwds)
                    x = cspc[cspc.find('(')+1:-1] # just the options() args
# cspc e.g.: option_kw("poly5","nearest","linear", default="poly5", comment="Interpolant (poly5,nearest,linear)")
                    x = x.split(',') # tokenize
                    # but! comment value may have commas in it, find it
                    # using it's equal sign, rm all after it
                    has_eq = [i for i in x if i.find('=')>=0]
                    if len(has_eq) > 0:
                        x = x[: x.index(has_eq[0]) ]
                    # rm spaces, extra quotes; rm kywd arg pairs
                    x = [i.strip("' ") for i in x if i.find('=')<0]
                    choicesOrMin = '|'+'|'.join(x)+'|' # IRAF format for enums
                elif chk_func_name.find('boolean') >= 0:     dtype = 'b'
                elif chk_func_name.find('float_or_') >= 0:   dtype = 'r'
                elif chk_func_name.find('float') >= 0:       dtype = 'R'
                elif chk_func_name.find('integer_or_') >= 0: dtype = 'i'
                elif chk_func_name.find('integer') >= 0:     dtype = 'I'
                elif chk_func_name.find('action') >= 0:      dtype = 'z'
                fields.append(dtype)
                fields.append('a')
                if type(val)==bool:
                    if val: fields.append('yes')
                    else:   fields.append('no')
                else:
                    fields.append(val)
                fields.append(choicesOrMin)
                fields.append(None)
                # Primarily use description from .cfgspc file (0). But, allow
                # overrides from .cfg file (1) if different.
                dscrp0 = chk_args_dict.get('comment','').strip() # ok if missing
                dscrp1 = cfgObj.inline_comments[key]
                if dscrp1 is None:
                    dscrp1 = ''

                while len(dscrp1) > 0 and dscrp1[0] in (' ','#'):
                    dscrp1 = dscrp1[1:] # .cfg file comments start with '#'

                dscrp1 = dscrp1.strip()
                # Now, decide what to do/say about the descriptions
                if len(dscrp1) > 0:
                    dscrp = dscrp0
                    if dscrp0 != dscrp1: # allow override if different
                        dscrp = dscrp1+eparoption.DSCRPTN_FLAG # flag it
                        if initialPass:
                            if dscrp0 == '' and cspc is None:
                                # this is a case where this par isn't in the
                                # .cfgspc; ignore, it is caught/error later
                                pass
                            else:
                                self.debug('Description of "'+key+ \
                                    '" overridden, from:  '+repr(dscrp0)+\
                                    '  to:  '+repr(dscrp1))
                    fields.append(dscrp)
                else:
                    # set the field for the GUI
                    fields.append(dscrp0)
                    # ALSO set it in the dict so it is written to file later
                    cfgObj.inline_comments[key] = '# '+dscrp0
                # This little section, while never intended to be used during
                # normal operation, could save a lot of manual work.
                if dumpCfgspcTo:
                    junk = cspc
                    junk = key+' = '+junk.strip()
                    if junk.find(' comment=')<0:
                        junk = junk[:-1]+", comment="+ \
                               repr(irafutils.stripQuotes(dscrp1.strip()))+")"
                    dumpCfgspcTo.write(junk+'\n')
                # Create the par
                if not toBeHidden or chk_func_name.find('action')==0:
                    par = basicpar.parFactory(fields, True)
                    par.setScope(scopePrefix)
                    retval.append(par)
                # else this is a hidden key

                # The next few items require a fully scoped name
                absKeyName = scopePrefix+'.'+key # assumed to be unique
                # Check for pars marked to be positional args
                if initialPass:
                    pos = chk_args_dict.get('pos')
                    if pos:
                        # we'll sort them later, on demand
                        self._posArgs.append( (int(pos), scopePrefix, key) )
                # Check for triggers and/or dependencies
                if initialPass:
                    # What triggers what? (thats why theres an 's' in the kwd)
                    # try "trigger" (old)
                    if chk_args_dict.get('trigger'):
                        print("WARNING: outdated version of .cfgspc!! for "+
                              self.__taskName+", 'trigger' unused for "+
                              absKeyName)
                    # try "triggers"
                    trgs = chk_args_dict.get('triggers')
                    if trgs and len(trgs)>0:
                        # eg. _allTriggers['STEP2.xy'] == ('_rule1_','_rule3_')
                        assert absKeyName not in self._allTriggers, \
                            'More than 1 of these in .cfgspc?: '+absKeyName
                        # we force this to always be a sequence
                        if isinstance(trgs, (list,tuple)):
                            self._allTriggers[absKeyName] = trgs
                        else:
                            self._allTriggers[absKeyName] = (trgs,)
                    # try "executes"
                    excs = chk_args_dict.get('executes')
                    if excs and len(excs)>0:
                        # eg. _allExecutes['STEP2.xy'] == ('_rule1_','_rule3_')
                        assert absKeyName not in self._allExecutes, \
                            'More than 1 of these in .cfgspc?: '+absKeyName
                        # we force this to always be a sequence
                        if isinstance(excs, (list,tuple)):
                            self._allExecutes[absKeyName] = excs
                        else:
                            self._allExecutes[absKeyName] = (excs,)

                    # Dependencies? (besides these used here, may someday
                    # add: 'range_from', 'warn_if', etc.)
                    depName = None
                    if not depName:
                        depType = 'active_if'
                        depName = chk_args_dict.get(depType) # e.g. =='_rule1_'
                    if not depName:
                        depType = 'inactive_if'
                        depName = chk_args_dict.get(depType)
                    if not depName:
                        depType = 'is_set_by'
                        depName = chk_args_dict.get(depType)
                    if not depName:
                        depType = 'set_yes_if'
                        depName = chk_args_dict.get(depType)
                    if not depName:
                        depType = 'set_no_if'
                        depName = chk_args_dict.get(depType)
                    if not depName:
                        depType = 'is_disabled_by'
                        depName = chk_args_dict.get(depType)
                    # NOTE - the above few lines stops at the first dependency
                    # found (depName) for a given par.  If, in the future a
                    # given par can have >1 dependency than we need to revamp!!
                    if depName:
                        # Add to _allDepdcs dict: (val is dict of pars:types)
                        #
                        # e.g. _allDepdcs['_rule1_'] == \
                        #        {'STEP3.ra':      'active_if',
                        #         'STEP3.dec':     'active_if',
                        #         'STEP3.azimuth': 'inactive_if'}
                        if depName in self._allDepdcs:
                            thisRulesDict = self._allDepdcs[depName]
                            assert not absKeyName in thisRulesDict, \
                                'Cant yet handle multiple actions for the '+ \
                                'same par and the same rule.  For "'+depName+ \
                                '" dict was: '+str(thisRulesDict)+ \
                                ' while trying to add to it: '+\
                                str({absKeyName:depType})
                            thisRulesDict[absKeyName] = depType
                        else:
                            self._allDepdcs[depName] = {absKeyName:depType}
                    # else no dependencies found for this chk_args_dict
        return retval


    def getTriggerStrings(self, parScope, parName):
        """ For a given item (scope + name), return all strings (in a tuple)
        that it is meant to trigger, if any exist.  Returns None is none. """
        # The data structure of _allTriggers was chosen for how easily/quickly
        # this particular access can be made here.
        fullName = parScope+'.'+parName
        return self._allTriggers.get(fullName) # returns None if unfound


    def getParsWhoDependOn(self, ruleName):
        """ Find any parameters which depend on the given trigger name. Returns
        None or a dict of {scopedName: dependencyName} from _allDepdcs. """
        # The data structure of _allDepdcs was chosen for how easily/quickly
        # this particular access can be made here.
        return self._allDepdcs.get(ruleName)


    def getExecuteStrings(self, parScope, parName):
        """ For a given item (scope + name), return all strings (in a tuple)
        that it is meant to execute, if any exist.  Returns None is none. """
        # The data structure of _allExecutes was chosen for how easily/quickly
        # this particular access can be made here.
        fullName = parScope+'.'+parName
        return self._allExecutes.get(fullName) # returns None if unfound


    def getPosArgs(self):
        """ Return a list, in order, of any parameters marked with "pos=N" in
            the .cfgspc file. """
        if len(self._posArgs) < 1: return []
        # The first item in the tuple is the index, so we now sort by it
        self._posArgs.sort()
        # Build a return list
        retval = []
        for idx, scope, name in self._posArgs:
            theDict, val = findScopedPar(self, scope, name)
            retval.append(val)
        return retval


    def getKwdArgs(self, flatten = False):
        """ Return a dict of all normal dict parameters - that is, all
            parameters NOT marked with "pos=N" in the .cfgspc file.  This will
            also exclude all hidden parameters (metadata, rules, etc). """

        # Start with a full deep-copy.  What complicates this method is the
        # idea of sub-sections.  This dict can have dicts as values, and so on.
        dcopy = self.dict() # ConfigObj docs say this is a deep-copy

        # First go through the dict removing all positional args
        for idx,scope,name in self._posArgs:
            theDict, val = findScopedPar(dcopy, scope, name)
            # 'theDict' may be dcopy, or it may be a dict under it
            theDict.pop(name)

        # Then go through the dict removing all hidden items ('_item_name_')
        for k in list(dcopy.keys()):
            if isHiddenName(k):
                dcopy.pop(k)

        # Done with the nominal operation
        if not flatten:
            return dcopy

        # They have asked us to flatten the structure - to bring all parameters
        # up to the top level, even if they are in sub-sections.  So we look
        # for values that are dicts.  We will throw something if we end up
        # with name collisions at the top level as a result of this.
        return flattenDictTree(dcopy)


    def canPerformValidation(self):
        """ Override this so we can do our own validation. tryValue() will
            be called as a result. """
        return True


    def knowAsNative(self):
        """ Override so we can keep native types in the internal dict. """
        return True


    def tryValue(self, name, val, scope=''):
        """ For the given item name (and scope), we are being asked to try
            the given value to see if it would pass validation.  We are not
            to set it, but just try it.  We return a tuple:
            If it fails, we return: (False,  the last known valid value).
            On success, we return: (True, None). """

        # SIMILARITY BETWEEN THIS AND setParam() SHOULD BE CONSOLIDATED!

        # Set the value, even if invalid.  It needs to be set before
        # the validation step (next).
        theDict, oldVal = findScopedPar(self, scope, name)
        if oldVal == val: return (True, None) # assume oldVal is valid
        theDict[name] = val

        # Check the proposed value.  Ideally, we'd like to
        # (somehow elegantly) only check this one item. For now, the best
        # shortcut is to only validate this section.
        ans=self.validate(self._vtor, preserve_errors=True, section=theDict)

        # No matter what ans is, immediately return the item to its original
        # value since we are only checking the value here - not setting.
        theDict[name] = oldVal

        # Now see what the validation check said
        errStr = ''
        if ans != True:
            flatStr = "All values are invalid!"
            if ans != False:
                flatStr = flattened2str(configobj.flatten_errors(self, ans))
            errStr = "Validation error: "+flatStr # for now this info is unused

        # Done
        if len(errStr): return (False, oldVal) # was an error
        else:           return (True, None)    # val is OK


    def listTheExtras(self, deleteAlso):
        """ Use ConfigObj's get_extra_values() call to find any extra/unknown
        parameters we may have loaded.  Return a string similar to findTheLost.
        If deleteAlso is True, this will also delete any extra/unknown items.
        """
        # get list of extras
        extras = configobj.get_extra_values(self)
        # extras is in format: [(sections, key), (sections, key), ]
        # but we need: [(sections, key, result), ...] - set all results to
        # a bool just to make it the right shape.  BUT, since we are in
        # here anyway, make that bool mean something - hide info in it about
        # whether that extra item is a section (1) or just a single par (0)
        #
        # simplified, this is:  expanded = [ (x+(abool,)) for x in extras]
        expanded = [ (x+ \
                       ( bool(len(x[0])<1 and hasattr(self[x[1]], 'keys')), ) \
                     ) for x in extras]
        retval = ''
        if expanded:
            retval = flattened2str(expanded, extra=1)
        # but before we return, delete them (from ourself!) if requested to
        if deleteAlso:
            for tup_to_del in extras:
                target = self
                # descend the tree to the dict where this items is located.
                # (this works because target is not a copy (because the dict
                #  type is mutable))
                location = tup_to_del[0]
                for subdict in location: target = target[subdict]
                # delete it
                target.pop(tup_to_del[1])

        return retval


# ---------------------------- helper functions --------------------------------


def findTheLost(config_file, configspec_file, skipHidden=True):
    """ Find any lost/missing parameters in this cfg file, compared to what
    the .cfgspc says should be there. This method is recommended by the
    ConfigObj docs. Return a stringified list of item errors. """
    # do some sanity checking, but don't (yet) make this a serious error
    if not os.path.exists(config_file):
        print("ERROR: Config file not found: "+config_file)
        return []
    if not os.path.exists(configspec_file):
        print("ERROR: Configspec file not found: "+configspec_file)
        return []
    tmpObj = configobj.ConfigObj(config_file, configspec=configspec_file)
    simval = configobj.SimpleVal()
    test = tmpObj.validate(simval)
    if test == True:
        return []
    # If we get here, there is a dict returned of {key1: bool, key2: bool}
    # which matches the shape of the config obj.  We need to walk it to
    # find the Falses, since they are the missing pars.
    missing = []
    flattened = configobj.flatten_errors(tmpObj, test)
    # But, before we move on, skip/eliminate any 'hidden' items from our list,
    # since hidden items are really supposed to be missing from the .cfg file.
    if len(flattened) > 0 and skipHidden:
        keepers = []
        for tup in flattened:
            keep = True
            # hidden section
            if len(tup[0])>0 and isHiddenName(tup[0][-1]):
                keep = False
            # hidden par (in a section, or at the top level)
            elif tup[1] is not None and isHiddenName(tup[1]):
                keep = False
            if keep:
                keepers.append(tup)
        flattened = keepers
    flatStr = flattened2str(flattened, missing=True)
    return flatStr


def isHiddenName(astr):
    """ Return True if this string name denotes a hidden par or section """
    if astr is not None and len(astr) > 2 and astr.startswith('_') and \
       astr.endswith('_'):
        return True
    else:
        return False


def flattened2str(flattened, missing=False, extra=False):
    """ Return a pretty-printed multi-line string version of the output of
    flatten_errors. Know that flattened comes in the form of a list
    of keys that failed. Each member of the list is a tuple::

        ([list of sections...], key, result)

    so we turn that into a string. Set missing to True if all the input
    problems are from missing items.  Set extra to True if all the input
    problems are from extra items. """

    if flattened is None or len(flattened) < 1:
        return ''
    retval = ''
    for sections, key, result in flattened:
        # Name the section and item, to start the message line
        if sections is None or len(sections) == 0:
            retval += '\t"'+key+'"'
        elif len(sections) == 1:
            if key is None:
                # a whole section is missing at the top-level; see if hidden
                junk = sections[0]
                if isHiddenName(junk):
                    continue # this missing or extra section is not an error
                else:
                    retval += '\tSection "'+sections[0]+'"'
            else:
                retval += '\t"'+sections[0]+'.'+key+'"'
        else: # len > 1
            joined = '.'.join(sections)
            joined = '"'+joined+'"'
            if key is None:
                retval +=  '\tSection '+joined
            else:
                retval +=  '\t"'+key+'" from '+joined
        # End the msg line with "what seems to be the trouble" with this one
        if missing and result==False:
            retval += ' is missing.'
        elif extra:
            if result:
                retval += ' is an unexpected section. Is your file out of date?'
            else:
                retval += ' is an unexpected parameter. Is your file out of date?'
        elif isinstance(result, bool):
            retval += ' has an invalid value'
        else:
            retval += ' is invalid, '+result.message
        retval += '\n\n'
    return retval.rstrip()



#!/usr/local/bin/python

""" This file holds our own over-rides for the standard Validator check
    functions.  We over-ride them so that we may add our own special keywords
    to them in the config_spec.

$Id$
"""
from __future__ import absolute_import, division, print_function # confidence high

from . import configobj, validate
from . import irafutils

STANDARD_KEYS = ['min', 'max', 'missing', 'default']
OVCDBG = False


def sigStrToKwArgsDict(checkFuncSig):
    """ Take a check function signature (string), and parse it to get a dict
        of the keyword args and their values. """
    p1 = checkFuncSig.find('(')
    p2 = checkFuncSig.rfind(')')
    assert p1 > 0 and p2 > 0 and p2 > p1, "Invalid signature: "+checkFuncSig
    argParts = irafutils.csvSplit(checkFuncSig[p1+1:p2], ',', True)
    argParts = [x.strip() for x in argParts]
    retval = {}
    for argPair in argParts:
        argSpl = argPair.split('=', 1)
        if len(argSpl) > 1:
            if argSpl[0] in retval:
                if isinstance(retval[argSpl[0]], (list,tuple)):
                    retval[argSpl[0]]+=(irafutils.stripQuotes(argSpl[1]),) # 3rd
                else: # 2nd in, so convert to tuple
                    retval[argSpl[0]] = (retval[argSpl[0]],
                                         irafutils.stripQuotes(argSpl[1]),)
            else:
                retval[argSpl[0]] = irafutils.stripQuotes(argSpl[1]) # 1st in
        else:
            retval[argSpl[0]] = None #  eg. found "triggers=, max=6, ..."
    return retval


def separateKeywords(kwArgsDict):
    """ Look through the keywords passed and separate the special ones we
        have added from the legal/standard ones.  Return both sets as two
        dicts (in a tuple), as (standardKws, ourKws) """
    standardKws = {}
    ourKws = {}
    for k in kwArgsDict:
        if k in STANDARD_KEYS:
            standardKws[k]=kwArgsDict[k]
        else:
            ourKws[k]=kwArgsDict[k]
    return (standardKws, ourKws)


def addKwdArgsToSig(sigStr, kwArgsDict):
    """ Alter the passed function signature string to add the given kewords """
    retval = sigStr
    if len(kwArgsDict) > 0:
        retval = retval.strip(' ,)') # open up the r.h.s. for more args
        for k in kwArgsDict:
            if retval[-1] != '(': retval += ", "
            retval += str(k)+"="+str(kwArgsDict[k])
        retval += ')'
    retval = retval
    return retval


def boolean_check_kw(val, *args, **kw):
    if OVCDBG: print("boolean_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    vtor = validate.Validator()
    checkFuncStr = "boolean"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def option_check_kw(val, *args, **kw):
    if OVCDBG: print("option_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    vtor = validate.Validator()
    checkFuncStr = "option"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def integer_check_kw(val, *args, **kw):
    if OVCDBG: print("integer_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    vtor = validate.Validator()
    checkFuncStr = "integer"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def integer_or_none_check_kw(val, *args, **kw):
    if OVCDBG: print("integer_or_none_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    if val in (None,'','None','NONE','INDEF'): return None # only difference
    vtor = validate.Validator()
    checkFuncStr = "integer"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def float_check_kw(val, *args, **kw):
    if OVCDBG: print("float_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    vtor = validate.Validator()
    checkFuncStr = "float"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def float_or_none_check_kw(val, *args, **kw):
    if OVCDBG: print("float_or_none_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    if val in (None,'','None','NONE','INDEF'): return None # only difference
    vtor = validate.Validator()
    checkFuncStr = "float"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


def string_check_kw(val, *args, **kw):
    if OVCDBG: print("string_kw for: "+str(val)+", args: "+str(args)+", kw: "+str(kw))
    vtor = validate.Validator()
    checkFuncStr = "string"+str(tuple(args))
    checkFuncStr = addKwdArgsToSig(checkFuncStr, separateKeywords(kw)[0])
    if OVCDBG: print("CFS: "+checkFuncStr+'\n')
    return vtor.check(checkFuncStr, val)


FUNC_DICT = {'boolean_kw':         boolean_check_kw,
             'option_kw':          option_check_kw,
             'integer_kw':         integer_check_kw,
             'integer_or_none_kw': integer_or_none_check_kw,
             'float_kw':           float_check_kw,
             'float_or_none_kw':   float_or_none_check_kw,
             'string_kw':          string_check_kw,
             'action_kw':          string_check_kw }



"""
Fit a line to a data set with optional weights.

Returns the parameters of the model, bo, b1:
Y = b0 + b1* X

:author: Nadia Dencheva
:version: '1.0 (2007-02-20)'

"""
from __future__ import absolute_import, division, print_function

import numpy as N

__version__ = '1.0'          # Release version number only
__vdate__ = '2007-02-20'     # Date of this version


def linefit(x, y, weights=None):

    """
    Parameters
    ----------
    y: 1D numpy array
        The data to be fitted
    x: 1D numpy array
        The x values of the y array. x and y must
        have the same shape.
    weights:   1D numpy array, must have the same shape as x and y
        weight values

    Examples
    --------
    >>> import numpy as N
    >>> from numpy.core import around
    >>> x = N.array([-5, -4 ,-3 ,-2 ,-1, 0, 1, 2, 3, 4, 5])
    >>> y = N.array([1, 5, 4, 7, 10, 8, 9, 13, 14, 13, 18])
    >>> around(linefit(x,y), decimals=5)
    array([9.27273, 1.43636])
    >>> x = N.array([1.3,1.3,2.0,2.0,2.7,3.3,3.3,3.7,3.7,4.,4.,4.,4.7,4.7,5.,5.3,5.3,5.3,5.7,6.,6.,6.3,6.7])
    >>> y = N.array([2.3,1.8,2.8,1.5,2.2,3.8,1.8,3.7,1.7,2.8,2.8,2.2,3.2,1.9,1.8,3.5,2.8,2.1,3.4,3.2,3.,3.,5.9])
    >>> around(linefit(x,y), decimals=5)
    array([1.42564, 0.31579])
    """
    if len(x) != len(y):
        print("Error: X and Y must have equal size\n")
        return
    n = len(x)
    w = N.zeros((n,n)).astype(N.float)
    if weights is None:
        for i in N.arange(n):
            w[i,i] = 1
    else:
        if len(weights) != n:
            print("Error: Weights must have the same size as X and Y.\n")
            return
        for i in N.arange(n):
            w[i,i] = weights[i]
    x = x.astype(N.float)
    y = y.astype(N.float)
    # take the weighted avg for calculatiing the covarince
    Xavg = N.sum(N.dot(w,x)) / N.sum(w.diagonal())
    Yavg = N.sum(N.dot(w,y)) / N.sum(w.diagonal())

    xm = x - Xavg
    xmt = N.transpose(xm)
    ym = y - Yavg

    b1 = N.dot(xmt,N.dot(w,ym)) / N.dot(xmt ,N.dot(w,xm))
    b0 = Yavg - b1 * Xavg

    return b0, b1



"""module irafglobals.py -- widely used IRAF constants and objects

NOTE!  This module does NOT require the installation of IRAF.  It's location
in stsci.tools is safe because it is intended to remain free of such dependency.

yes, no         Boolean values
IrafError       Standard IRAF exception
Verbose         Flag indicating verbosity level
userIrafHome    User's IRAF home directory (./ or ~/iraf/)
userWorkingHome User's working home directory (the directory
                when this module gets imported.)
EOF             End-of-file indicator object
INDEF           Undefined object
IrafTask        "Tag" class for IrafTask type.
IrafPkg         "Tag" class for IrafPkg type

This is defined so it is safe to say 'from irafglobals import *'

The tag classes do nothing except allow checks of types via (e.g.)
isinstance(o,IrafTask).  Including it here decouples the other classes
from the module that actually implements IrafTask, greatly reducing the
need for mutual imports of modules by one another.

$Id$

Taken from pyraf.irafglobals, originally signed "R. White, 2000 Jan 5"
"""
from __future__ import absolute_import, division

import sys
PY3K = sys.version_info[0] > 2

if PY3K:
    string_types = str
    number_types = (int, float)
else:
    string_types = basestring
    number_types = (int, long, float)

import os
from . import compmixin
_os = os
_compmixin = compmixin
del os, compmixin

class IrafError(Exception):
    def __init__(self, msg, errno=-1, errmsg="", errtask=""):
        Exception.__init__(self, msg)
        self.errno = errno
        self.errmsg = errmsg or msg
        self.errtask = errtask

# -----------------------------------------------------
# Verbose: verbosity flag
# -----------------------------------------------------

# make Verbose an instance of a class so it can be imported
# into other modules and changed by them

class _VerboseClass(_compmixin.ComparableIntBaseMixin):
    """Container class for verbosity (or other) value"""
    def __init__(self, value=0): self.value = value
    def set(self, value): self.value = value
    def get(self): return self.value
    def _cmpkey(self): return self.value
    def __nonzero__(self): return self.value != 0
    def __bool__(self):    return self.value != 0
    def __str__(self): return str(self.value)

Verbose = _VerboseClass()

# -----------------------------------------------------
# userWorkingHome is current working directory
# -----------------------------------------------------

userWorkingHome = _os.getcwd()

# -----------------------------------------------------
# userIrafHome is location of user's IRAF home directory
# -----------------------------------------------------

# If login.cl exists here, use this directory as home.
# Otherwise look for ~/iraf.

if _os.path.exists('./login.cl'):
    userIrafHome = _os.path.join(userWorkingHome,'')
elif _os.path.exists(_os.path.expanduser('~/.iraf/login.cl')):
    userIrafHome = _os.path.expanduser('~/.iraf')
else:
    userIrafHome = _os.path.join(_os.getenv('HOME','.'),'iraf','')
    if not _os.path.exists(userIrafHome):
        # no ~/iraf, just use '.' as home
        userIrafHome = _os.path.join(userWorkingHome,'')

# -----------------------------------------------------
# Boolean constant class
# -----------------------------------------------------

class _Boolean(_compmixin.ComparableMixin):
    """Class of boolean constant object"""
    def __init__(self, value=None):
        # change value to 1 or 0
        if value:
            self.__value = 1
        else:
            self.__value = 0
        self.__strvalue = ["no", "yes"][self.__value]

    def __copy__(self):
        """Don't bother to make a copy"""
        return self

    def __deepcopy__(self, memo=None):
        """Don't bother to make a copy"""
        return self

    def _compare(self, other, method):
        # _Boolean vs. _Boolean
        if isinstance(other, _Boolean):
            return method(self.__value, other.__value)
        # _Boolean vs. string:
        # If a string, compare with string value of this parameter.
        # Allow uppercase "YES", "NO" as well as lowercase.
        # Also allows single letter abbrevation "y" or "n".
        if isinstance(other, string_types):
            ovalue = other.lower()
            if len(ovalue)==1:
                return method(self.__strvalue[0], ovalue)
            else:
                return method(self.__strvalue, ovalue)
        # _Boolean vs. all other types (int, float, bool, etc) - treat this
        # value like an integer
        return method(self.__value, other)

    def __nonzero__(self): return self.__value != 0
    def __bool__(self):    return self.__value != 0
    def __repr__(self): return self.__strvalue
    def __str__(self): return self.__strvalue
    def __int__(self): return self.__value
    def __float__(self): return float(self.__value)

# create yes, no boolean values

yes = _Boolean(1)
no = _Boolean(0)


# -----------------------------------------------------
# define end-of-file object
# if printed, says 'EOF'
# if converted to integer, has value -2 (special IRAF value)
# Implemented as a singleton, although the singleton
# nature is not really essential
# -----------------------------------------------------

class _EOFClass(_compmixin.ComparableMixin):
    """Class of singleton EOF (end-of-file) object"""
    def __init__(self):
        global EOF
        if EOF is not None:
            # only allow one to be created
            raise RuntimeError("Use EOF object, not _EOFClass")

    def __copy__(self):
        """Not allowed to make a copy"""
        return self

    def __deepcopy__(self, memo=None):
        """Not allowed to make a copy"""
        return self

    def _compare(self, other, method):
        if isinstance(other, _EOFClass):
            # Despite trying to create only one EOF object, there
            # could be more than one.  All EOFs are equal.
            return method(1, 1)
        if isinstance(other, string_types):
            # If a string, compare with 'EOF'
            return method("EOF", other)
        if isinstance(other, number_types):
            # If a number, compare with -2
            return method(-2, other)
        # what else could it be?
        return NotImplemented

    def __repr__(self): return "EOF"
    def __str__(self): return "EOF"
    def __int__(self): return -2
    def __float__(self): return -2.0

# initialize EOF to None first so singleton scheme works

EOF = None
EOF = _EOFClass()

# -----------------------------------------------------
# define IRAF-like INDEF object
# -----------------------------------------------------

class _INDEFClass(object):
    """Class of singleton INDEF (undefined) object"""

    def __new__(cls):
        # Guido's example Singleton pattern
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        # this use of super gets the correct version of __new__ for the
        # int and float subclasses too
        cls.__it__ = it = super(_INDEFClass, cls).__new__(cls)
        return it

    def __copy__(self):
        """Not allowed to make a copy"""
        return self

    def __deepcopy__(self, memo=None):
        """Not allowed to make a copy"""
        return self

    def __lt__(self, other): return INDEF
    def __le__(self, other): return INDEF
    def __gt__(self, other): return INDEF
    def __ge__(self, other): return INDEF

    def __eq__(self, other):
        # Despite trying to create only one INDEF object, there
        # could be more than one.  All INDEFs are equal.
        # Also allow "INDEF" - CDS 17Nov2011
        return isinstance(other, _INDEFClass) or (other and str(other)=="INDEF")

    def __ne__(self, other):
        return not self.__eq__(other)

    def __repr__(self): return "INDEF"
    def __str__(self): return "INDEF"

    __oct__ = __str__
    __hex__ = __str__

    # type conversions return various types of INDEF objects
    # this is necessary for Python 2.4

    def __int__(self): return _INDEF_int
    def __long__(self): return _INDEF_int
    def __float__(self): return _INDEF_float

    def __nonzero__(self): return False # need bool return type

    # all operations on INDEF return INDEF

    def __add__(self, other): return INDEF

    __sub__    = __add__
    __mul__    = __add__
    __rmul__   = __add__
    __div__    = __add__
    __mod__    = __add__
    __divmod__ = __add__
    __pow__    = __add__
    __lshift__ = __add__
    __rshift__ = __add__
    __and__    = __add__
    __xor__    = __add__
    __or__     = __add__

    __radd__    = __add__
    __rsub__    = __add__
    __rmul__    = __add__
    __rrmul__   = __add__
    __rdiv__    = __add__
    __rmod__    = __add__
    __rdivmod__ = __add__
    __rpow__    = __add__
    __rlshift__ = __add__
    __rrshift__ = __add__
    __rand__    = __add__
    __rxor__    = __add__
    __ror__     = __add__

    def __neg__(self): return INDEF

    __pos__    = __neg__
    __abs__    = __neg__
    __invert__ = __neg__

INDEF = _INDEFClass()

# Classes that inherit from built-in types are required for Python 2.4
# so that int and float conversion functions work correctly.
# Unfortunately, if you call int(_INDEF_int) it ignores the
# __int__ method and returns zero, so these objects should be
# used sparingly and replaced with standard INDEF whereever
# possible.

class _INDEFClass_int(_INDEFClass, int): pass
class _INDEFClass_float(_INDEFClass, float): pass
_INDEF_int = _INDEFClass_int()
_INDEF_float = _INDEFClass_float()

# -----------------------------------------------------
# define IRAF-like EPSILON object
# -----------------------------------------------------

class _EPSILONClass(_compmixin.ComparableFloatBaseMixin):
    """Class of singleton EPSILON object, for floating-point comparison"""

    def __new__(cls):
        # Guido's example Singleton pattern
        it = cls.__dict__.get("__it__")
        if it is not None:
            return it
        cls.__it__ = it = super(_EPSILONClass, cls).__new__(cls)
        return it

    def __init__(self):
        self.__dict__["_value"] = None

    def setvalue(self):
        DEFAULT_VALUE = 1.192e-7
        hlib = _os.environ.get("hlib")
        if hlib is None:
            self._value = DEFAULT_VALUE
            return
        fd = open(_os.path.join(hlib, "mach.h"))
        lines = fd.readlines()
        fd.close()
        foundit = 0
        for line in lines:
            words = line.split()
            if len(words) < 1 or words[0] == "#":
                continue
            if words[0] == "define" and words[1] == "EPSILONR":
                strvalue = words[2]
                if strvalue[0] == "(":
                    strvalue = strvalue[1:-1]
                self._value = float(strvalue)
                foundit = 1
                break
        if not foundit:
            self._value = DEFAULT_VALUE

    def __copy__(self):
        """Not allowed to make a copy"""
        return self

    def __deepcopy__(self, memo=None):
        """Not allowed to make a copy"""
        return self

    def __setattr__(self, name, value):
        """Not allowed to modify the value or add a new attribute"""
        if name == "_value":
            if self.__dict__["_value"] is None:
                self.__dict__["_value"] = value
            else:
                raise RuntimeError("epsilon cannot be modified")
        else:
            pass

    def __delattr__(self, value):
        """Not allowed to delete the value"""
        pass

    def _cmpkey(self): return self._value

    def __repr__(self): return "%.6g" % self._value
    def __str__(self): return "%.6g" % self._value

    __oct__ = None
    __hex__ = None

    def __int__(self): return 0
    def __long__(self): return 0
    def __float__(self): return self._value

    def __nonzero__(self): return True # need bool return type

    def __add__(self, other):
        return self._value + other

    def __sub__(self, other):
        return self._value - other

    def __mul__(self, other):
        return self._value * other

    def __div__(self, other):
        return self._value / other

    def __mod__(self, other):
        return self._value % other

    def __divmod__(self, other):
        return (self._value // other, self._value % other)

    def __pow__(self, other):
        return self._value ** other

    def __neg__(self):
        return -self._value

    def __pos__(self):
        return self._value

    def __abs__(self):
        return abs(self._value)

    # arguments in reverse order
    def __radd__(self, other):
        return other + self._value

    def __rsub__(self, other):
        return other - self._value

    def __rmul__(self, other):
        return other * self._value

    def __rdiv__(self, other):
        return other / self._value

    def __rmod__(self, other):
        return other % self._value

    def __rdivmod__(self, other):
        return (other // self._value, other % self._value)

    def __rpow__(self, other):
        return other ** self._value

epsilon = _EPSILONClass()
epsilon.setvalue()

# -----------------------------------------------------
# tag classes
# -----------------------------------------------------

class IrafTask:
    pass

class IrafPkg(IrafTask):
    pass



#!/usr/bin/env python
import os
import pkgutil
import sys
from setuptools import setup, find_packages
from subprocess import check_call, CalledProcessError


if not pkgutil.find_loader('relic'):
    relic_local = os.path.exists('relic')
    relic_submodule = (relic_local and
                       os.path.exists('.gitmodules') and
                       not os.listdir('relic'))
    try:
        if relic_submodule:
            check_call(['git', 'submodule', 'update', '--init', '--recursive'])
        elif not relic_local:
            check_call(['git', 'clone', 'https://github.com/spacetelescope/relic.git'])

        sys.path.insert(1, 'relic')
    except CalledProcessError as e:
        print(e)
        exit(1)

import relic.release

version = relic.release.get_info()
relic.release.write_template(version, 'lib/stsci/tools')

setup(
    name = 'stsci.tools',
    version = version.pep386,
    author = 'STScI',
    author_email = 'help@stsci.edu',
    description = 'Collection of STScI utility functions',
    url = 'https://github.com/spacetelescope/stsci.tools',
    classifiers = [
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Topic :: Scientific/Engineering :: Astronomy',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
    install_requires = [
        'astropy',
        'numpy',
    ],
    setup_requires = [
        'pytest-runner'
    ],
    tests_require = [
        'pytest',
        'pytest-doctestplus'
    ],
    package_dir = {
        '': 'lib',
    },
    packages = find_packages('lib'),
    package_data = {
        '': ['LICENSE.txt'],
        'stsci/tools/tests': ['data/*.*']
    },
    entry_points = {
        'console_scripts': [
            'convertwaiveredfits=stsci.tools.convertwaiveredfits:main',
            'convertlog=stsci.tools.convertlog:main'
        ],
    },
)



from __future__ import absolute_import

import sys

import pytest

from ..compmixin import ComparableMixin

if sys.version_info[0] < 3:
    string_types = basestring
else:
    string_types = str


class SimpleStr(ComparableMixin):
    def __init__(self, v):
        self.val = str(v)  # all input turned to string

    def __str__(self):
        return str(self.val)

    def _cmpkey(self):
        return self.val


class AnyType(ComparableMixin):
    def __init__(self, v):
        self.val = v  # leave all input typed as is

    def __str__(self):
        return str(self.val)

    # define this instead of _cmpkey - handle ALL sorts of scenarios,
    # except intentionally don't compare self strings (strlen>1) with integers
    # so we have a case which fails in our test below
    def _compare(self, other, method):
        if isinstance(other, self.__class__):
            # recurse, get 2 logic below
            return self._compare(other.val, method)
        if isinstance(other, string_types):
            return method(str(self.val), other)
        elif other is None and self.val is None:
            return method(0, 0)
        elif other is None:
            # coerce to str compare
            return method(str(self.val), '')
        elif isinstance(other, int):
            # handle ONLY case where self.val is a single char or an int
            if isinstance(self.val, string_types) and len(self.val) == 1:
                return method(ord(self.val), other)
            else:
                # assume we are int-like
                return method(int(self.val), other)
        try:
            return method(self.val, other)
        except (AttributeError, TypeError):
            return NotImplemented


def test_SimpleStr():
    a = SimpleStr('a')
    b = SimpleStr('b')
    c = SimpleStr('c')
    two = SimpleStr(2)

    # compare two SimpleStr objects
    assert str(a > b) == "False"
    assert str(a < b) == "True"
    assert str(a <= b) == "True"
    assert str(a == b) == "False"
    assert str(b == b) == "True"
    assert str(a < c) == "True"
    assert str(a <= c) == "True"
    assert str(a != c) == "True"
    assert str(c != c) == "False"
    assert str(c == c) == "True"
    assert str(b < two) == "False"
    assert str(b >= two) == "True"
    assert str(b == two) == "False"
    assert str([str(jj) for jj in sorted([b, a, two, c])]
               ) == "['2', 'a', 'b', 'c']"


def test_AnyType():
    x = AnyType('x')
    y = AnyType('yyy')
    z = AnyType(0)
    nn = AnyType(None)

    # compare two AnyType objects
    assert str(x > y) == "False"
    assert str(x < y) == "True"
    assert str(x <= y) == "True"
    assert str(x == y) == "False"
    assert str(y == y) == "True"
    assert str(x < z) == "False"
    assert str(x <= z) == "False"
    assert str(x > z) == "True"
    assert str(x != z) == "True"
    assert str(z != z) == "False"
    assert str(z == z) == "True"
    assert str(y < nn) == "False"
    assert str(y >= nn) == "True"
    assert str(y == nn) == "False"
    assert str(nn == nn) == "True"
    assert str([str(jj) for jj in sorted([y, x, nn, z])]
               ) == "['None', '0', 'x', 'yyy']"

    # compare AnyType objects to built-in types
    assert str(x < 0) == "False"
    assert str(x <= 0) == "False"
    assert str(x > 0) == "True"
    assert str(x != 0) == "True"
    assert str(x == 0) == "False"
    assert str(x < None) == "False"
    assert str(x <= None) == "False"
    assert str(x > None) == "True"
    assert str(x != None) == "True"
    assert str(x == None) == "False"
    assert str(x < "abc") == "False"
    assert str(x <= "abc") == "False"
    assert str(x > "abc") == "True"
    assert str(x != "abc") == "True"
    assert str(x == "abc") == "False"
    assert str(y < None) == "False"
    assert str(y <= None) == "False"
    assert str(y > None) == "True"
    assert str(y != None) == "True"
    assert str(y == None) == "False"
    assert str(y < "abc") == "False"
    assert str(y <= "abc") == "False"
    assert str(y > "abc") == "True"
    assert str(y != "abc") == "True"
    assert str(y == "abc") == "False"


def test_raise_on_integer_comparison():
    y = AnyType('yyy')
    z = AnyType(0)
    with pytest.raises(ValueError):
        y == z  # AnyType intentionally doesn't compare strlen>1 to ints


def test_raise_on_sort():
    y = AnyType('yyy')
    z = AnyType(0)
    with pytest.raises(ValueError):
        sorted([z, y])



#!/usr/bin/env python

# $Id$

"""
        readgeis: Read GEIS file and convert it to a FITS extension file.

        License: http://www.stsci.edu/resources/software_hardware/pyraf/LICENSE

        Usage:

                readgeis.py [options] GEISname FITSname

                GEISname is the input GEIS file in GEIS format, and FITSname
                is the output file in FITS format. GEISname can be a
                directory name.  In this case, it will try to use all `*.??h`
                files as input file names.

                If FITSname is omitted or is a directory name, this task will
                try to construct the output names from the input names, i.e.:

                abc.xyh will have an output name of abc_xyf.fits

        :Options:

        -h     print the help (this text)

        :Example:

        If used in Pythons script, a user can, e. g.::

            >>> from stsci.tools import readgeis
            >>> hdulist = readgeis.readgeis(GEISFileName)  # doctest: +SKIP
            (do whatever with hdulist)
            >>> hdulist.writeto(FITSFileName)  # doctest: +SKIP

        The most basic usage from the command line::

            readgeis.py test1.hhh test1.fits

        This command will convert the input GEIS file test1.hhh to
        a FITS file test1.fits.


        From the command line::

            readgeis.py .

        this will convert all `*.??h` files in the current directory
        to FITS files (of corresponding names) and write them in the
        current directory.


        Another example of usage from the command line::

            readgeis.py "u*" "*"

        this will convert all `u*.??h` files in the current directory
        to FITS files (of corresponding names) and write them in the
        current directory.  Note that when using wild cards, it is
        necessary to put them in quotes.

"""

# Developed by Science Software Branch, STScI, USA.
# This version needs pyfits 0.9.6.3 or later
# and numpy version 1.0.4 or later

from __future__ import division, print_function # confidence high

__version__ = "2.2 (18 Feb, 2011), \xa9 AURA"

import os, sys
from astropy.io import fits
import numpy
from numpy import memmap
from functools import reduce

def stsci(hdulist):
    """For STScI GEIS files, need to do extra steps."""

    instrument = hdulist[0].header.get('INSTRUME', '')

    # Update extension header keywords
    if instrument in ("WFPC2", "FOC"):
        rootname = hdulist[0].header.get('ROOTNAME', '')
        filetype = hdulist[0].header.get('FILETYPE', '')
        for i in range(1, len(hdulist)):
            # Add name and extver attributes to match PyFITS data structure
            hdulist[i].name = filetype
            hdulist[i]._extver = i
            # Add extension keywords for this chip to extension
            hdulist[i].header['EXPNAME'] = (rootname, "9 character exposure identifier")
            hdulist[i].header['EXTVER']= (i, "extension version number")
            hdulist[i].header['EXTNAME'] = (filetype, "extension name")
            hdulist[i].header['INHERIT'] = (True, "inherit the primary header")
            hdulist[i].header['ROOTNAME'] = (rootname, "rootname of the observation set")


def stsci2(hdulist, filename):
    """For STScI GEIS files, need to do extra steps."""

    # Write output file name to the primary header
    instrument = hdulist[0].header.get('INSTRUME', '')
    if instrument in ("WFPC2", "FOC"):
        hdulist[0].header['FILENAME'] = filename


def readgeis(input):

    """Input GEIS files "input" will be read and a HDUList object will
       be returned.

       The user can use the writeto method to write the HDUList object to
       a FITS file.
    """

    global dat
    cardLen = fits.Card.length

    # input file(s) must be of the form *.??h and *.??d
    if input[-1] != 'h' or input[-4] != '.':
        raise "Illegal input GEIS file name %s" % input

    data_file = input[:-1]+'d'

    _os = sys.platform
    if _os[:5] == 'linux' or _os[:5] == 'win32' or _os[:5] == 'sunos' or _os[:3] == 'osf' or _os[:6] == 'darwin':
        bytes_per_line = cardLen+1
    else:
        raise "Platform %s is not supported (yet)." % _os

    geis_fmt = {'REAL':'f', 'INTEGER':'i', 'LOGICAL':'i','CHARACTER':'S'}
    end_card = 'END'+' '* (cardLen-3)

    # open input file
    im = open(input)

    # Generate the primary HDU
    cards = []
    while 1:
        line = im.read(bytes_per_line)[:cardLen]
        line = line[:8].upper() + line[8:]
        if line == end_card:
            break
        cards.append(fits.Card.fromstring(line))

    phdr = fits.Header(cards)
    im.close()

    _naxis0 = phdr.get('NAXIS', 0)
    _naxis = [phdr['NAXIS'+str(j)] for j in range(1, _naxis0+1)]
    _naxis.insert(0, _naxis0)
    _bitpix = phdr['BITPIX']
    _psize = phdr['PSIZE']
    if phdr['DATATYPE'][:4] == 'REAL':
        _bitpix = -_bitpix
    if _naxis0 > 0:
        size = reduce(lambda x,y:x*y, _naxis[1:])
        data_size = abs(_bitpix) * size // 8
    else:
        data_size = 0
    group_size = data_size + _psize // 8

    # decode the group parameter definitions,
    # group parameters will become extension header
    groups = phdr['GROUPS']
    gcount = phdr['GCOUNT']
    pcount = phdr['PCOUNT']

    formats = []
    bools = []
    floats = []
    _range = range(1, pcount+1)
    key = [phdr['PTYPE'+str(j)] for j in _range]
    comm = [phdr.cards['PTYPE'+str(j)].comment for j in _range]

    # delete group parameter definition header keywords
    _list = ['PTYPE'+str(j) for j in _range] + \
            ['PDTYPE'+str(j) for j in _range] + \
            ['PSIZE'+str(j) for j in _range] + \
            ['DATATYPE', 'PSIZE', 'GCOUNT', 'PCOUNT', 'BSCALE', 'BZERO']

    # Construct record array formats for the group parameters
    # as interpreted from the Primary header file
    for i in range(1, pcount+1):
        ptype = key[i-1]
        pdtype = phdr['PDTYPE'+str(i)]
        star = pdtype.find('*')
        _type = pdtype[:star]
        _bytes = pdtype[star+1:]

        # collect boolean keywords since they need special attention later

        if _type == 'LOGICAL':
            bools.append(i)
        if pdtype == 'REAL*4':
            floats.append(i)

        fmt = geis_fmt[_type] + _bytes
        formats.append((ptype,fmt))

    _shape = _naxis[1:]
    _shape.reverse()
    _code = fits.BITPIX2DTYPE[_bitpix]
    _bscale = phdr.get('BSCALE', 1)
    _bzero = phdr.get('BZERO', 0)
    if phdr['DATATYPE'][:10] == 'UNSIGNED*2':
        _uint16 = 1
        _bzero = 32768
    else:
        _uint16 = 0

    # delete from the end, so it will not conflict with previous delete
    for i in range(len(phdr)-1, -1, -1):
        if phdr.cards[i].keyword in _list:
            del phdr[i]

    # clean up other primary header keywords
    phdr['SIMPLE'] = True
    phdr['BITPIX'] = 16
    phdr['GROUPS'] = False
    _after = 'NAXIS'
    if _naxis0 > 0:
        _after += str(_naxis0)
    phdr.set('EXTEND', value=True, comment="FITS dataset may contain extensions", after=_after)
    phdr.set('NEXTEND', value=gcount, comment="Number of standard extensions")

    hdulist = fits.HDUList([fits.PrimaryHDU(header=phdr, data=None)])

    # Use copy-on-write for all data types since byteswap may be needed
    # in some platforms.
    f1 = open(data_file, mode='rb')
    dat = f1.read()
#    dat = memmap(data_file, mode='c')
    hdulist.mmobject = dat

    errormsg = ""

    loc = 0
    for k in range(gcount):
        ext_dat = numpy.fromstring(dat[loc:loc+data_size], dtype=_code)
        ext_dat = ext_dat.reshape(_shape)
        if _uint16:
            ext_dat += _bzero
        # Check to see whether there are any NaN's or infs which might indicate
        # a byte-swapping problem, such as being written out on little-endian
        #   and being read in on big-endian or vice-versa.
        if _code.find('float') >= 0 and \
            (numpy.any(numpy.isnan(ext_dat)) or numpy.any(numpy.isinf(ext_dat))):
            errormsg += "===================================\n"
            errormsg += "= WARNING:                        =\n"
            errormsg += "=  Input image:                   =\n"
            errormsg += input+"[%d]\n"%(k+1)
            errormsg += "=  had floating point data values =\n"
            errormsg += "=  of NaN and/or Inf.             =\n"
            errormsg += "===================================\n"
        elif _code.find('int') >= 0:
            # Check INT data for max values
            ext_dat_frac,ext_dat_exp = numpy.frexp(ext_dat)
            if ext_dat_exp.max() == int(_bitpix) - 1:
                # Potential problems with byteswapping
                errormsg += "===================================\n"
                errormsg += "= WARNING:                        =\n"
                errormsg += "=  Input image:                   =\n"
                errormsg += input+"[%d]\n"%(k+1)
                errormsg += "=  had integer data values        =\n"
                errormsg += "=  with maximum bitvalues.        =\n"
                errormsg += "===================================\n"

        ext_hdu = fits.ImageHDU(data=ext_dat)

        rec = numpy.fromstring(dat[loc+data_size:loc+group_size], dtype=formats)

        loc += group_size

        # Create separate PyFITS Card objects for each entry in 'rec'
        for i in range(1, pcount+1):
            #val = rec.field(i-1)[0]
            val = rec[0][i-1]
            if val.dtype.kind == 'S':
                val = val.decode('ascii')

            if i in bools:
                if val:
                    val = True
                else:
                    val = False

            if i in floats:
                # use fromstring, format in Card is deprecated in pyfits 0.9
                _str = '%-8s= %20.7G / %s' % (key[i-1], val, comm[i-1])
                _card = fits.Card.fromstring(_str)
            else:
                _card = fits.Card(keyword=key[i-1], value=val, comment=comm[i-1])

            ext_hdu.header.append(_card)

        # deal with bscale/bzero
        if (_bscale != 1 or _bzero != 0):
            ext_hdu.header['BSCALE'] = _bscale
            ext_hdu.header['BZERO'] = _bzero

        hdulist.append(ext_hdu)

    if errormsg != "":
        errormsg += "===================================\n"
        errormsg += "=  This file may have been        =\n"
        errormsg += "=  written out on a platform      =\n"
        errormsg += "=  with a different byte-order.   =\n"
        errormsg += "=                                 =\n"
        errormsg += "=  Please verify that the values  =\n"
        errormsg += "=  are correct or apply the       =\n"
        errormsg += "=  '.byteswap()' method.          =\n"
        errormsg += "===================================\n"
        print(errormsg)

    f1.close()
    stsci(hdulist)
    return hdulist

def parse_path(f1, f2):

    """Parse two input arguments and return two lists of file names"""

    import glob

    # if second argument is missing or is a wild card, point it
    # to the current directory
    f2 = f2.strip()
    if f2 == '' or f2 == '*':
        f2 = './'

    # if the first argument is a directory, use all GEIS files
    if os.path.isdir(f1):
        f1 = os.path.join(f1, '*.??h')
    list1 = glob.glob(f1)
    list1 = [name for name in list1 if name[-1] == 'h' and name[-4] == '.']

    # if the second argument is a directory, use file names in the
    # first argument to construct file names, i.e.
    # abc.xyh will be converted to abc_xyf.fits
    if os.path.isdir(f2):
        list2 = []
        for file in list1:
            name = os.path.split(file)[-1]
            fitsname = name[:-4] + '_' + name[-3:-1] + 'f.fits'
            list2.append(os.path.join(f2, fitsname))
    else:
        list2 = [s.strip() for s in f2.split(",")]

    if list1 == [] or list2 == []:
        err_msg = ""
        if list1 == []:
            err_msg += "Input files `{:s}` not usable/available. ".format(f1)

        if list2 == []:
            err_msg += "Input files `{:s}` not usable/available. ".format(f2)

        raise IOError(err_msg)

    else:
        return list1, list2

#-------------------------------------------------------------------------------
# special initialization when this is the main program

if __name__ == "__main__":

    import getopt

    try:
        optlist, args = getopt.getopt(sys.argv[1:], 'h')
    except getopt.error as e:
        print(str(e))
        print(__doc__)
        print("\t", __version__)

    # initialize default values
    help = 0

    # read options
    for opt, value in optlist:
        if opt == "-h":
            help = 1

    if (help):
        print(__doc__)
        print("\t", __version__)
    else:
        if len(args) == 1:
            args.append('')
        list1, list2 = parse_path (args[0], args[1])
        npairs = min (len(list1), len(list2))
        for i in range(npairs):
            if os.path.exists(list2[i]):
                print("Output file %s already exists, skip." % list2[i])
                break
            try:
                hdulist = readgeis(list1[i])
                stsci2(hdulist, list2[i])
                hdulist.writeto(list2[i])
                hdulist.close()
                print("%s -> %s" % (list1[i], list2[i]))
            except Exception as e :
                print("Conversion fails for %s: %s" % (list1[i], str(e)))
                break

"""

Copyright (C) 2003 Association of Universities for Research in Astronomy (AURA)

Redistribution and use in source and binary forms, with or without
modification, are permitted provided that the following conditions are met:

    1. Redistributions of source code must retain the above copyright
      notice, this list of conditions and the following disclaimer.

    2. Redistributions in binary form must reproduce the above
      copyright notice, this list of conditions and the following
      disclaimer in the documentation and/or other materials provided
      with the distribution.

    3. The name of AURA and its representatives may not be used to
      endorse or promote products derived from this software without
      specific prior written permission.

THIS SOFTWARE IS PROVIDED BY AURA ``AS IS'' AND ANY EXPRESS OR IMPLIED
WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF
MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
DISCLAIMED. IN NO EVENT SHALL AURA BE LIABLE FOR ANY DIRECT, INDIRECT,
INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,
BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS
OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND
ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR
TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE
USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH
DAMAGE.
"""
